# coding: utf-8


def get_group_1_and_2(tsv, contrast):
    """
        Based on the TSV file given in parameter, compute indexes of each group
    Args:
        tsv: (str) path to the tsv file containing information on subjects/sessions with all covariates
        contrast: (str) name of a column of the tsv

    Returns:
        first_group_idx: (list of int) list of indexes of first group
        second_group_idx: (list of int) list of indexes of second group
        class_names: (list of str of len 2) list of the class names read in the column contrast of the tsv
    """
    import pandas as pds
    from clinica.utils.exceptions import ClinicaException

    # StatisticsVolume pipeline has been instantiated with tsv_file=tsv,
    # so check of existence and integrity have succeeded
    # No further check are done when trying to read it
    tsv = pds.read_csv(tsv, sep='\t')
    columns = list(tsv.columns)

    # An error is raised if the contrast column is not found in the tsv
    if contrast not in columns:
        raise ClinicaException(contrast + ' is not present in ' + tsv)

    # list(set(my_list)) gives unique values of my_list
    class_names = list(set(tsv[contrast]))

    # This is a 2-sample t-test: we can only allow 2 classes
    if len(class_names) != 2:
        raise ClinicaException('It must exist only 2 classes in the column ' + contrast
                               + ' to perform 2-sample t-tests. Here Clinica found: ' + str(class_names))
    first_group_idx = [i for i, label in enumerate(list(tsv[contrast])) if label == class_names[0]]
    second_group_idx = [i for i, label in enumerate(list(tsv[contrast])) if label == class_names[1]]

    return first_group_idx, second_group_idx, class_names


def model_creation(tsv, contrast, idx_group1, idx_group2, file_list, template_file):
    """
        Create the matlab .m file for the instantiation of the 2-sample t-test model in SPM

    Args:
        tsv: (str) path to the tsv file containing information on subjects/sessions with all covariates
        contrast: (str) name of a column of the tsv
        idx_group1: (list of int) list of indexes of first group
        idx_group2: (list of int) list of indexes of second group
        file_list: List of files used in the statistical test. Their order is the same as it appears on the tsv file
        template_file: (str) path to the template file used to generate the .m file

    Returns:
        current_model: (str) path to the matlab files with all the @TEXT replaced with the correct names
        covariates: list of str with the names of covariates
    """
    from os.path import join, dirname, isfile, abspath, isdir
    from shutil import rmtree
    from clinica.utils.exceptions import ClinicaException
    from numbers import Number
    from os import remove, mkdir
    import numpy as np
    import pandas as pds
    import clinica.pipelines.statistics_volume.statistics_volume_utils as utls

    # Get template for model creation
    if not isfile(template_file):
        raise RuntimeError('[Error] ' + template_file + ' could not be found !')

    # Create current model filename
    current_model = abspath('./current_model_creation.m')

    if isfile(current_model):
        # This should never be reached
        remove(current_model)

    # Read template
    with open(template_file, 'r') as file:
        filedata = file.read()

    # Create output folder for SPM (and remove it if it already exists)
    output_folder = abspath(join(dirname(current_model), '..', '2_sample_t_test'))
    if isdir(output_folder):
        rmtree(output_folder)
    mkdir(output_folder)

    # Replace string in matlab file to set the output directory, and all the scans used for group 1 and 2
    filedata = filedata.replace('@OUTPUTDIR', '\'' + output_folder + '\'')
    filedata = filedata.replace('@SCANS1', utls.unravel_list_for_matlab([f for i, f in enumerate(file_list) if i in idx_group1]))
    filedata = filedata.replace('@SCANS2', utls.unravel_list_for_matlab([f for i, f in enumerate(file_list) if i in idx_group2]))

    # Write filedata in the output file
    with open(current_model, 'w+') as file:
        file.write(filedata)

    # Add our covariates
    tsv = pds.read_csv(tsv, sep='\t')
    columns_stripped = [elem.strip(' ') for elem in list(tsv.columns)]
    if columns_stripped != list(tsv.columns):
        raise ClinicaException('[Error] Check the column of your tsv file ' + tsv
                               + 'Whitespace in the column names can cause errors')
    covariates = [elem for elem in columns_stripped if elem not in ['participant_id', 'session_id', contrast]]
    for covar_number, covar in enumerate(covariates, start=1):
        current_covar_data = list(tsv[covar])
        if isinstance(current_covar_data[0], str):
            # Transform data
            temp_data = [elem.replace(',', '.') for elem in current_covar_data]
            if all(utls.is_number(elem) for elem in temp_data):
                current_covar_data = [float(elem) for elem in temp_data]
            else:
                # categorical variables (like Male; Female; M, F etc...)
                unique_values = list(np.unique(np.array(current_covar_data)))
                current_covar_data = [unique_values.index(elem) for elem in current_covar_data]

        elif isinstance(current_covar_data[0], Number):
            # Do nothing
            pass

        current_covar_data_group1 = [elem for i, elem in enumerate(current_covar_data) if i in idx_group1]
        current_covar_data_group2 = [elem for i, elem in enumerate(current_covar_data) if i in idx_group2]
        covar_data_concatenated = current_covar_data_group1 + current_covar_data_group2
        utls.write_covariate_lines(current_model, covar_number, covar, covar_data_concatenated)

    # Tell matlab to run the script at the end
    with open(current_model, 'a') as file:
        file.write('spm_jobman(\'run\', matlabbatch)')
    return current_model, covariates


def is_number(s: str):
    """

    Args:
        s: (str) string to test if it can be converted into float

    Returns:
        True or False
    """
    try:
        # Try the conversion, if it is not possible, error will be raised
        float(s)
        return True
    except ValueError:
        return False


def unravel_list_for_matlab(my_list):
    """
        Unpack a list into a Matlab compliant format to insert in .m files.
    Args:
        my_list: (list) of str

    Returns:
        (str) that join the different str of the list, with '
    """
    result = '\',\''.join(my_list)
    result = '\'' + result + '\''
    return result


def write_covariate_lines(m_file_to_write_in, covar_number, covar_name, covar_values):
    """
        Use this function to add covariate lines in the Matlab file m_file_to_write_in for one covariate
    Args:
        m_file_to_write_in: (str) path to the m-file
        covar_number: (int) this is the number of the covariate (must start at 1)
        covar_name: (str) name of the covariate (ex: 'age', 'sex')
        covar_values: (list) of float with the values of the covariates

    Returns:
        nothing
    """
    from os.path import isfile

    assert isfile(m_file_to_write_in), 'Could not find file ' + m_file_to_write_in
    covar_values_string = [str(elem) for elem in covar_values]
    covar_values_for_script = '[' + ' '.join(covar_values_string) + ']'

    m_file = open(m_file_to_write_in, 'a')
    m_file.write('matlabbatch{1}.spm.stats.factorial_design.cov(' + str(covar_number) + ').c = ' + covar_values_for_script + ';\n')
    m_file.write('matlabbatch{1}.spm.stats.factorial_design.cov(' + str(covar_number) + ').cname = \'' + covar_name + '\';\n')
    m_file.write('matlabbatch{1}.spm.stats.factorial_design.cov(' + str(covar_number) + ').iCFI = 1;\n')
    m_file.write('matlabbatch{1}.spm.stats.factorial_design.cov(' + str(covar_number) + ').iCC = 1;\n')
    m_file.close()


def run_m_script(m_file):
    """
        Runs a matlab m file for SPM, determining automatically if it must be launched with SPM or SPM Standalone
        If launch with spm standalone, the line 'spm_jobman('run', matlabbatch)' must be removed because unnecessary

    Args:
        m_file: (str) path to Matlab m file

    Returns:
        output_mat_file: (str) path to the SPM.mat file needed in SPM analysis
    """
    from os.path import isfile, dirname, basename, abspath, join
    from os import system
    from clinica.utils.spm import spm_standalone_is_available
    import clinica.pipelines.statistics_volume.statistics_volume_utils as utls
    from nipype.interfaces.matlab import MatlabCommand, get_matlab_command
    import platform

    assert isinstance(m_file, str), '[Error] Argument must be a string'
    if not isfile(m_file):
        raise FileNotFoundError('[Error] File ' + m_file + 'does not exist')
    assert m_file[-2:] == '.m', '[Error] ' + m_file + ' is not a Matlab file (extension must be .m)'

    # Generate command line to run
    if spm_standalone_is_available():
        utls.delete_last_line(m_file)
        # SPM standalone must be run directly from its root folder
        if platform.system().lower().startswith('darwin'):
            # Mac OS
            cmdline = 'cd $SPMSTANDALONE_HOME && ./run_spm12.sh $MCR_HOME batch ' + m_file
        elif platform.system().lower().startswith('linux'):
            # Linux OS
            cmdline = '$SPMSTANDALONE_HOME/run_spm12.sh $MCR_HOME batch ' + m_file
        else:
            raise SystemError('Clinica only support Mac OS and Linux')
        system(cmdline)
    else:
        MatlabCommand.set_default_matlab_cmd(get_matlab_command())
        matlab = MatlabCommand()
        if platform.system().lower().startswith('linux'):
            matlab.inputs.args = '-nosoftwareopengl'
        matlab.inputs.paths = dirname(m_file)
        matlab.inputs.script = basename(m_file)[:-2]
        matlab.inputs.single_comp_thread = False
        matlab.inputs.logfile = abspath('./matlab_output.log')
        matlab.run()
    output_mat_file = abspath(join(dirname(m_file), '..', '2_sample_t_test', 'SPM.mat'))
    if not isfile(output_mat_file):
        raise RuntimeError('Output matrix ' + output_mat_file + ' was not produced')
    return output_mat_file


def delete_last_line(filename):
    """
    Use this function to remove the call to spm jobman if m file is used with SPM standalone
    Args:
        filename: path to filename

    Returns:
        Nothing
    """
    import os
    with open(filename, "r+", encoding="utf-8") as file:

        # Move the pointer (similar to a cursor in a text editor) to the end of the file
        file.seek(0, os.SEEK_END)

        # This code means the following code skips the very last character in the file -
        # i.e. in the case the last line is null we delete the last line
        # and the penultimate one
        pos = file.tell() - 1

        # Read each character in the file one at a time from the penultimate
        # character going backwards, searching for a newline character
        # If we find a new line, exit the search
        while pos > 0 and file.read(1) != "\n":
            pos -= 1
            file.seek(pos, os.SEEK_SET)

        # So long as we're not at the start of the file, delete all the characters ahead
        # of this position
        if pos > 0:
            file.seek(pos, os.SEEK_SET)
            file.truncate()


def estimate(mat_file, template_file):
    """
        Make a copy of the template file (for estimation) and replace @SPMMAT by the real path to SPM.mat (mat_file)
    Args:
        mat_file: (str) path to the SPM.mat file of the SPM analysis
        template_file: (str) path to the template file for the estimation of the model

    Returns:
        current_model_estimation: (str) path to the template file filled with SPM.mat information, ready to be launched
    """
    from os.path import abspath

    # Read template
    with open(template_file, 'r') as file:
        filedata = file.read()
    # Replace by the real path to spm.mat
    filedata = filedata.replace('@SPMMAT', '\'' + mat_file + '\'')
    current_model_estimation = abspath('./current_model_creation.m')
    with open(current_model_estimation, 'w+') as file:
        file.write(filedata)

    return current_model_estimation


def results(mat_file, template_file, method, threshold):
    """
        Make a copy of the template file (for results) and replace @SPMMAT by the real path to SPM.mat (mat_file)
    Args:
        mat_file: (str) path to the SPM.mat file of the SPM analysis
        template_file: (str) path to the template file for getting the results of the model
        method: (str)
        threshold

    Returns:
        current_model_estimation: (str) path to the template file filled with SPM.mat information, ready to be launched
    """
    from os.path import abspath
    import time

    # Read template
    with open(template_file, 'r') as file:
        filedata = file.read()
    # Replace by the real path to spm.mat
    filedata = filedata.replace('@SPMMAT', '\'' + mat_file + '\'')
    filedata = filedata.replace('@CORRECTIONMETHOD', '\'' + method + '\'')
    filedata = filedata.replace('@THRESH', str(threshold))
    if method != 'none':
        # Necessary so that the 2 computations does not overlap
        time.sleep(15)
    current_model_result = abspath('./current_model_result.m')
    with open(current_model_result, 'w+') as file:
        file.write(filedata)

    return current_model_result


def contrast(mat_file, template_file, covariates, class_names):
    """
        Make a copy of the template file (for results) and replace @SPMMAT, @COVARNUMBER, @GROUP1, @GROUP2
        by the corresponding variables

    Args:
        mat_file: (str) path to the SPM.mat file of the SPM analysis
        template_file: (str) path to the template file for getting the results of the model
        covariates: (list) of str: list of covariates
        class_names: (list) of str of length 2 that correspond to the 2 classes for the group comparison

    Returns:
        current_model_estimation: (str) path to the template file filled with variables, ready to be launched
    """
    from os.path import abspath

    number_of_covariates = len(covariates)

    # Read template
    with open(template_file, 'r') as file:
        filedata = file.read()
    # Replace by the real path to spm.mat
    filedata = filedata.replace('@SPMMAT', '\'' + mat_file + '\'')
    filedata = filedata.replace('@COVARNUMBER', '0 ' * number_of_covariates)
    filedata = filedata.replace('@GROUP1', '\'' + class_names[0] + '\'')
    filedata = filedata.replace('@GROUP2', '\'' + class_names[1] + '\'')
    current_model_estimation = abspath('./current_model_contrast.m')
    with open(current_model_estimation, 'w+') as file:
        file.write(filedata)

    return current_model_estimation


def read_output(spm_mat, class_names, covariates, group_label, fwhm, measure):
    """
        Once analysis is done, grab all the different filenames and rename them in current directory according to class
        names
    Args:
        spm_mat: (str) path to the SPM.mat file of the SPM analysis
        class_names: (list) of str of length 2 that correspond to the 2 classes for the group comparison
        covariates: (list) of str: list of covariates
        group_label: name of the group label
        fwhm: fwhm in mm used
        measure: measure used

    Returns:
        spmT_0001: (str) path to t maps for the first group comparison
        spmT_0002: (str) path to t maps for the second group comparison
        spm_figures: (list) path to figure files
        variance_of_error: (str) path to variance of error
        resels_per_voxels: (str) path to resels per voxel
        mask: (str) path to mask of included voxels
        regression_coeff: (str list) path to regression coefficients
        contrasts: (str list) path to weighted parameter estimation for the 2 contrasts

    """
    from os.path import join, dirname, isdir, isfile, abspath
    from os import listdir
    from shutil import copyfile

    if not isfile(spm_mat):
        if not isdir(dirname(spm_mat)):
            raise RuntimeError('[Error] output folder ' + dirname(spm_mat) + ' does not exist.')
        else:
            raise RuntimeError('[Error] SPM matrix ' + spm_mat + ' does not exist.')

    list_files = [f for f in listdir(dirname(spm_mat)) if not f.startswith('.')]

    # Handle figure files
    figures = [abspath(join(dirname(spm_mat), f)) for f in list_files if f.endswith('png')]
    if len(figures) < 2:
        raise RuntimeError('[Error] Figures were not generated')
    fig_number = [int(f[-7:-4]) for f in figures]
    spm_figures = [abspath('./' + 'group-' + group_label + '_report-' + str(i) + '.png') for i in fig_number]
    for old_name, new_name in zip(figures, spm_figures):
        copyfile(old_name, new_name)

    # Handle spm t maps
    spm_T = [abspath(join(dirname(spm_mat), f)) for f in list_files if f.startswith('spmT')]
    spm_T = sorted(spm_T)
    if len(spm_T) != 2:
        raise RuntimeError('[Error] ' + str(len(spm_T)) + ' SPM t-map(s) were found')
    if fwhm:
        spmT_0001 = abspath('group-' + group_label + '_' + class_names[0] + '-lt-' + class_names[1] + '_measure-' + measure + '_fwhm-' + str(int(fwhm)) + '_TStatistics.nii')
        spmT_0002 = abspath('group-' + group_label + '_' + class_names[1] + '-lt-' + class_names[0] + '_measure-' + measure + '_fwhm-' + str(int(fwhm)) + '_TStatistics.nii')
    else:
        spmT_0001 = abspath('group-' + group_label + '_' + class_names[0] + '-lt-' + class_names[1] + '_measure-' + measure + '_TStatistics.nii')
        spmT_0002 = abspath('group-' + group_label + '_' + class_names[1] + '-lt-' + class_names[0] + '_measure-' + measure + '_TStatistics.nii')
    copyfile(join(dirname(spm_mat), 'spmT_0001.nii'), spmT_0001)
    copyfile(join(dirname(spm_mat), 'spmT_0002.nii'), spmT_0002)

    variance_of_error = abspath('./group-' + group_label + '_VarianceError.nii')
    copyfile(abspath(join(dirname(spm_mat), 'ResMS.nii')), variance_of_error)

    resels_per_voxels = abspath('./resels_per_voxel.nii')
    copyfile(abspath(join(dirname(spm_mat), 'RPV.nii')), resels_per_voxels)

    mask = abspath('./included_voxel_mask.nii')
    copyfile(abspath(join(dirname(spm_mat), 'mask.nii')), mask)

    # Handle beta files
    betas = [abspath(join(dirname(spm_mat), f)) for f in list_files if f.startswith('beta_')]
    betas = sorted(betas)
    if len(betas) != 2 + len(covariates):
        raise RuntimeError('[Error] Not enough betas files found in output directory')
    regression_coeff_covar = [abspath('./' + covar + '.nii') for covar in covariates]
    regression_coeff = [abspath('./' + class_names[0] + '.nii'),
                        abspath('./' + class_names[1] + '.nii')]
    regression_coeff.extend(regression_coeff_covar)
    # Order is respected:
    for beta, reg_coeff in zip(betas, regression_coeff):
        copyfile(beta, reg_coeff)

    # Handle contrast files
    con_files = [abspath(join(dirname(spm_mat), f)) for f in list_files if f.startswith('con_')]
    if len(con_files) != 2:
        raise RuntimeError('There must exists only 2 contrast files !')
    contrasts = [abspath('group-' + group_label + '_' + class_names[0] + '-lt-' + class_names[1] + '_measure-' + measure + '_contrast.nii'),
                 abspath('group-' + group_label + '_' + class_names[1] + '-lt-' + class_names[0] + '_measure-' + measure + '_contrast.nii')]
    for con, contrast in zip(con_files, contrasts):
        copyfile(con, contrast)

    return spmT_0001, spmT_0002, spm_figures, variance_of_error, resels_per_voxels, mask, regression_coeff, contrasts



# coding: utf8


class Descriptor(object):
    """
    Class that handles the conversion of a raw image name from the NIFD dataset,
        handles the BIDS format and the hierarchy of quality
        (T1 images with DIS3D will be selected over basic T1 images,
        the descriptor in charge of finding T1 images with DIS3D will have a higher priority value than the descriptor in charge of finding images with just T1 in its name)
    A descriptor object is created from a dictionnary written in the config_dcm2bids.json file
    To add a descriptor to the descriptors list that is used in convert_images, add it to the config_dcm2bids.json file
    """

    def __init__(self, dic_des):
        self.dataType = None
        self.modalityLabel = None
        self.customLabels = None
        self.priority = 0

        self.criteria = False
        self.Modality = None
        self.SeriesDescription = None

        if "dataType" in dic_des:
            self.dataType = dic_des["dataType"]
        if "modalityLabel" in dic_des:
            self.modalityLabel = dic_des["modalityLabel"]
        if "customLabels" in dic_des:
            self.customLabels = dic_des["customLabels"]
        if "priority" in dic_des:
            self.priority = int(dic_des["priority"])

        # Criteria attributes
        if "criteria" in dic_des:
            self.criteria = True
            if "Modality" in dic_des["criteria"]:
                self.Modality = dic_des["criteria"]["Modality"]
            if "SeriesDescription" in dic_des["criteria"]:
                self.SeriesDescription = dic_des["criteria"]["SeriesDescription"]

    def describes(self, str_image):
        """
        Uses self.SeriesDescription to determine whether self describes str_image
        For instance, if self.SeriesDescription = "*FLAIR*^*DIS3D:*flair*^*DIS3D"
        every image containing the string "FLAIR" and ending with "DIS3D" will make this method return True
        an image containing the string "flair" and ending with "DIS3D" will also return True
        '*' stands for any string
        '^' stands for AND
        ':' stands for OR

        Args:
            str_image: Image name

        Returns:
            Boolean: True if str_image corresponds to the pattern written in self.SeriesDescription, else False
        """

        def test_desc(str_image, desc):
            if desc[0] == '*':
                if desc[-1] == '*':
                    return desc[1:-1] in str_image
                else:
                    return str_image[-len(desc) + 1:] == desc[1:]

            else:
                if desc[-1] == '*':
                    return str_image[:len(desc) - 1] == desc[:-1]
                else:
                    return str_image == desc

        descriptions = self.SeriesDescription.split(':')
        for desc in descriptions:
            sub_desc = desc.split('^')
            bool = True
            for sub in sub_desc:
                bool = bool and test_desc(str_image, sub)

            if bool:
                return True
        return False

    def get_bids_info(self):
        s = ""
        if self.customLabels is not None:
            s += self.customLabels + '_'
        if self.modalityLabel is not None:
            s += self.modalityLabel + '_'

        if len(s) == 0:
            return s
        return s[:-1]

    def __str__(self):
        s = ""

        if self.dataType is not None:
            s += "dataType : " + self.dataType + ", "
        if self.modalityLabel is not None:
            s += "modalityLabel : " + self.modalityLabel + ", "
        if self.customLabels is not None:
            s += "customLabels : " + self.customLabels + ", "
        if self.priority != 0:
            s += "priority : " + str(self.priority) + ", "

        # Criteria attributes
        if self.criteria:
            s += "\n(criteria) "
            if self.Modality is not None:
                s += "Modality : " + self.Modality + ", "
            if self.SeriesDescription is not None:
                s += "SeriesDescription : " + self.SeriesDescription + ", "

        if len(s) > 2:
            return s[:-2]
        return s



# coding: utf8

import clinica.engine as ce


class StatisticsSurfaceCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 'statistics-surface'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Surface-based mass-univariate analysis with SurfStat:\n'
                             'http://clinica.run/doc/Pipelines/Stats_Surface/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        from colorama import Fore

        # Clinica compulsory arguments
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_label",
                                  help='User-defined identifier for the provided group of subjects.')
        clinica_comp.add_argument("orig_input_data",
                                  help='Type of surface-based feature: type '
                                  '\'t1-freesurfer\' to use cortical thickness, '
                                  '\'pet-surface\' to use projected FDG-PET data or '
                                  '\'custom-pipeline\' to use you own data in CAPS directory '
                                  '(see Wiki for details).',
                                  choices=['t1-freesurfer', 'pet-surface', 'custom-pipeline'])
        clinica_comp.add_argument("glm_type",
                                  help='String based on the GLM type for the hypothesis. '
                                  'You can choose between \'group_comparison\' and \'correlation\'.',
                                  choices=['group_comparison', 'correlation'],
                                  )
        clinica_comp.add_argument("subject_visits_with_covariates_tsv",
                                  help='TSV file containing a list of subjects with their sessions and all '
                                       'the covariates and factors needed for the GLM.')
        clinica_comp.add_argument("contrast",
                                  help='String to define the contrast matrix for the GLM, e.g. group. Please note '
                                       'that, when you want to perform negative correlation, the sign is ignored '
                                       'by the command line.')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-c", "--covariates",
                              type=str, default=None, metavar='\'covariate_1 ... covariate_n\'',
                              help='Covariates of the form --covariates \'covariate_1 ... covariate_n\'. '
                                   'Each covariate must match the column name of the TSV file. '
                                   'On default, no covariates are taken.')
        optional.add_argument("-fwhm", "--full_width_at_half_maximum",
                              type=int, default=20,
                              help='FWHM for the surface smoothing '
                                   '(default: --full_width_at_half_maximum %(default)s).')
        # Optional arguments for custom pipeline
        opt_custom_input = self._args.add_argument_group(
            '%sPipeline options if you selected custom-pipeline%s' % (Fore.BLUE, Fore.RESET)
        )
        opt_custom_input.add_argument("-cf", "--custom_file",
                                      type=str, default=None,
                                      help='Pattern of file inside CAPS directory using @subject, @session, '
                                           '@fwhm, @hemi. '
                                           'This flag must be specified with the --measure_label flag). '
                                           'See Wiki for an example.')
        opt_custom_input.add_argument("-ml", "--measure_label",
                                      type=str, default=None,
                                      help='Name of the feature type, it will be saved on the CAPS '
                                           '_measure-FEATURE_LABEL key-value association. '
                                           'This flag must be specified with the --custom_file flag). '
                                           'See Wiki for an example.')
        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments(add_tsv_flag=False)
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-ct", "--cluster_threshold",
                              type=float, default=0.001,
                              help='Threshold to define a cluster in the process of cluster-wise correction '
                                   '(default: --cluster_threshold %(default)s).')

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .statistics_surface_pipeline import StatisticsSurface
        from .statistics_surface_utils import (get_t1_freesurfer_custom_file,
                                               get_fdg_pet_surface_custom_file)
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit
        from clinica.utils.exceptions import ClinicaException

        # FreeSurfer cortical thickness
        if args.orig_input_data == 't1-freesurfer':
            args.custom_file = get_t1_freesurfer_custom_file()
            args.measure_label = 'ct'
        # PET cortical projection
        elif args.orig_input_data == 'pet-surface':
            args.custom_file = get_fdg_pet_surface_custom_file()
            args.measure_label = 'fdg'
        else:
            if (args.custom_file is None) or (args.measure_label is None):
                raise ClinicaException('You must set --measure_label and --custom_file flags.')

        parameters = {
            'orig_input_data': args.orig_input_data,
            'group_label': args.group_label,
            'covariates': args.covariates,
            'contrast': args.contrast,
            'glm_type': args.glm_type,
            'custom_file': args.custom_file,
            'measure_label': args.measure_label,
            'full_width_at_half_maximum': args.full_width_at_half_maximum,
            'cluster_threshold': args.cluster_threshold,
        }
        pipeline = StatisticsSurface(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subject_visits_with_covariates_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8

# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
from nipype import config

import clinica.pipelines.engine as cpe

cfg = dict(execution={'parameterize_dirs': False})
config.update_config(cfg)


class T1FreeSurfer(cpe.Pipeline):
    """FreeSurfer-based processing of T1-weighted MR images.

    Returns:
        A clinica pipeline object containing the T1FreeSurfer pipeline.
    """

    @staticmethod
    def get_processed_images(caps_directory, subjects, sessions):
        import os
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import T1_FS_DESTRIEUX
        from clinica.utils.filemanip import extract_image_ids
        image_ids = []
        if os.path.isdir(caps_directory):
            t1_freesurfer_files = clinica_file_reader(
                subjects, sessions,
                caps_directory, T1_FS_DESTRIEUX, False
            )
            image_ids = extract_image_ids(t1_freesurfer_files)
        return image_ids

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from colorama import Fore
        from clinica.utils.stream import cprint

        if "-dontrun" in self.parameters['recon_all_args'].split(' '):
            cprint('%s[Warning] Found -dontrun flag for FreeSurfer recon-all. '
                   'Please note that this will not run the segmentation.%s' %
                   (Fore.YELLOW, Fore.RESET))

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurfer pipeline is:
                * t1w (str): Path of the T1 weighted image in BIDS format

        Returns:
            A list of (string) input fields name.
        """
        return ['t1w']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Note:
            The list of outputs of the T1FreeSurfer pipeline is:
                * image_id (str): Image ID (e.g. sub-CLNC01_ses-M00)

        Returns:
            A list of (string) output fields name.
        """
        return ['image_id']

    def build_input_node(self):
        """Build and connect an input node to the pipeline.

        Raise:
            ClinicaBIDSError: If there are duplicated files or missing files for any subject
        """
        import os
        from colorama import Fore
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.iotools.utils.data_handling import check_volume_location_in_world_coordinate_system
        from clinica.utils.exceptions import ClinicaBIDSError, ClinicaException
        from clinica.utils.filemanip import extract_subjects_sessions_from_filename, save_participants_sessions
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import T1W_NII
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        # Display image(s) already present in CAPS folder
        # ===============================================
        processed_ids = self.get_processed_images(self.caps_directory, self.subjects, self.sessions)
        if len(processed_ids) > 0:
            cprint("%sClinica found %s image(s) already processed in CAPS directory:%s" %
                   (Fore.YELLOW, len(processed_ids), Fore.RESET))
            for image_id in processed_ids:
                cprint("%s\t%s%s" % (Fore.YELLOW, image_id.replace('_', ' | '), Fore.RESET))
            if self.overwrite_caps:
                output_folder = "<CAPS>/subjects/<participant_id>/<session_id>/t1/freesurfer_cross_sectional"
                cprint("%s\nOutput folders in %s will be recreated.\n%s" % (Fore.YELLOW, output_folder, Fore.RESET))
            else:
                cprint("%s\nImage(s) will be ignored by Clinica.\n%s" % (Fore.YELLOW, Fore.RESET))
                input_ids = [p_id + '_' + s_id for p_id, s_id in
                             zip(self.subjects, self.sessions)]
                to_process_ids = list(set(input_ids) - set(processed_ids))
                self.subjects, self.sessions = extract_subjects_sessions_from_filename(to_process_ids)

        # Inputs from anat/ folder
        # ========================
        # T1w file:
        try:
            t1w_files = clinica_file_reader(self.subjects,
                                            self.sessions,
                                            self.bids_directory,
                                            T1W_NII)
        except ClinicaException as e:
            err_msg = 'Clinica faced error(s) while trying to read files in your BIDS directory.\n' + str(e)
            raise ClinicaBIDSError(err_msg)

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        save_participants_sessions(self.subjects, self.sessions, folder_participants_tsv)

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint('List available in %s' % os.path.join(folder_participants_tsv, 'participants.tsv'))
            cprint('The pipeline will last approximately 10 hours per image.')

        read_node = npe.Node(name="ReadingFiles",
                             iterables=[
                                 ('t1w', t1w_files),
                             ],
                             synchronize=True,
                             interface=nutil.IdentityInterface(
                                 fields=self.get_input_fields())
                             )
        check_volume_location_in_world_coordinate_system(t1w_files, self.bids_directory)
        self.connect([
            (read_node, self.input_node, [('t1w', 't1w')]),
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from .t1_freesurfer_utils import save_to_caps
        import os

        save_to_caps = npe.Node(interface=nutil.Function(
            input_names=['source_dir', 'image_id', 'caps_dir', 'overwrite_caps'],
            output_names=['image_id'],
            function=save_to_caps),
            name='SaveToCaps')
        save_to_caps.inputs.source_dir = os.path.join(self.base_dir, self.name, 'ReconAll')
        save_to_caps.inputs.caps_dir = self.caps_directory
        save_to_caps.inputs.overwrite_caps = self.overwrite_caps

        self.connect([
            (self.output_node, save_to_caps, [('image_id', 'image_id')]),
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import os
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.freesurfer.preprocess import ReconAll
        from .t1_freesurfer_utils import (init_input_node, write_tsv_files)

        # Nodes declaration
        # =================
        # Initialize the pipeline
        #   - Extract <image_id> (e.g. sub-CLNC01_ses-M00) T1w filename;
        #   - Check FOV of T1w;
        #   - Create <subjects_dir> folder in <WD>/<Pipeline.name>/ReconAll/<image_id>/;
        #   - Print begin execution message.
        init_input = npe.Node(
            interface=nutil.Function(
                input_names=['t1w', 'recon_all_args', 'output_dir'],
                output_names=['image_id', 't1w', 'flags', 'subjects_dir'],
                function=init_input_node),
            name='0-InitPipeline')
        init_input.inputs.recon_all_args = self.parameters['recon_all_args']
        init_input.inputs.output_dir = os.path.join(self.base_dir, self.name, 'ReconAll')

        # Run recon-all command
        # FreeSurfer segmentation will be in <subjects_dir>/<image_id>/
        recon_all = npe.Node(interface=ReconAll(),
                             name='1-SegmentationReconAll')
        recon_all.inputs.directive = 'all'

        # Generate TSV files containing a summary of the regional statistics
        # in <subjects_dir>/regional_measures
        create_tsv = npe.Node(interface=nutil.Function(
            input_names=['subjects_dir', 'image_id'],
            output_names=['image_id'],
            function=write_tsv_files),
            name='2-CreateTsvFiles')

        # Connections
        # ===========
        self.connect([
            # Get <image_id> from input_node and print begin message
            (self.input_node, init_input, [('t1w', 't1w')]),
            # Run recon-all command
            (init_input, recon_all, [('subjects_dir', 'subjects_dir'),
                                     ('t1w', 'T1_files'),
                                     ('image_id', 'subject_id'),
                                     ('flags',  'flags')]),
            # Generate TSV files
            (init_input, create_tsv, [('subjects_dir', 'subjects_dir')]),
            (recon_all,  create_tsv, [('subject_id', 'image_id')]),
            # Output node
            (create_tsv, self.output_node, [('image_id', 'image_id')]),
        ])



# coding: utf-8

"""
Module for converting PIB PET of ADNI
"""


def convert_adni_pib_pet(source_dir, csv_dir, dest_dir, subjs_list=None):
    """Convert PIB PET images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of PIB PET images. Output will be stored in %s.' % path.join(dest_dir, 'conversion_info'))
    images = compute_pib_pet_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of PIB PET images found. Exporting images into BIDS ...')
    paths_to_bids(images, dest_dir, 'pib')
    cprint(Fore.GREEN + 'PIB PET conversion done.' + Fore.RESET)


def compute_pib_pet_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """Compute the paths to the PIB PET images and store them in a tsv file

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: a dataframe with all the paths to the PET images that will be converted into BIDS

    """

    import pandas as pd
    import os
    from os import path
    from clinica.iotools.converters.adni_to_bids.adni_utils import get_images_pet, find_image_path

    pet_pib_col = ['Phase', 'Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date', 'Study_ID',
                   'Series_ID', 'Image_ID', 'Original']
    pet_pib_df = pd.DataFrame(columns=pet_pib_col)
    pet_pib_dfs_list = []

    # Loading needed .csv files
    pibqc = pd.read_csv(path.join(csv_dir, 'PIBQC.csv'), sep=',', low_memory=False)
    pet_meta_list = pd.read_csv(path.join(csv_dir, 'PET_META_LIST.csv'), sep=',', low_memory=False)

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list['Subject'] == subj]

        if subject_pet_meta.empty:
            continue

        # QC for PIB PET images
        pet_qc_subj = pibqc[(pibqc.PASS == 1) & (pibqc.RID == int(subj[-4:]))]

        sequences_preprocessing_step = ['PIB Co-registered, Averaged']
        subj_dfs_list = get_images_pet(subj, pet_qc_subj, subject_pet_meta, pet_pib_col, 'PIB-PET',
                                       sequences_preprocessing_step, viscode_field="VISCODE")
        if subj_dfs_list:
            pet_pib_dfs_list += subj_dfs_list

    if pet_pib_dfs_list:
        pet_pib_df = pd.concat(pet_pib_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = []

    # Removing known exceptions from images to convert
    if not pet_pib_df.empty:
        error_ind = pet_pib_df.index[pet_pib_df.apply(lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors),
                                                      axis=1)]
        pet_pib_df.drop(error_ind, inplace=True)

    images = find_image_path(pet_pib_df, source_dir, 'PIB', 'I', 'Image_ID')

    pib_csv_path = path.join(dest_dir, 'conversion_info')
    if not os.path.exists(pib_csv_path):
        os.mkdir(pib_csv_path)
    images.to_csv(path.join(pib_csv_path, 'pib_pet_paths.tsv'), sep='\t', index=False)

    return images



# coding: utf8

import clinica.engine as ce


class T1LinearCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline.
        """

        self._name = 't1-linear'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = ('Affine registration of T1w images to the MNI standard space:\n'
                             'http://clinica.run/doc/Pipelines/T1_Linear/')

    def define_options(self):
        """Define the sub-command arguments
        """
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES

        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id...)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')

        # Clinica optional arguments
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])

        optional.add_argument("-ui", "--uncropped_image",
                              help='''Do not crop the image with template
                              (cropped image are suggested for using with DL
                              models)''',
                              action='store_true',
                              default=False)

        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .t1_linear_pipeline import T1Linear
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
                'uncropped_image': args.uncropped_image
        }

        # Most of the time, you will want to instantiate your pipeline with a
        # BIDS and CAPS directory as inputs:
        pipeline = T1Linear(
             bids_directory=self.absolute_path(args.bids_directory),
             caps_directory=self.absolute_path(args.caps_directory),
             tsv_file=self.absolute_path(args.subjects_sessions_tsv),
             base_dir=self.absolute_path(args.working_directory),
             parameters=parameters,
             name=self.name
             )

        if args.n_procs:
            exec_pipeline = pipeline.run(
                    plugin='MultiProc',
                    plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8

import clinica.engine as ce


class StatisticsVolumeCorrectionCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 'statistics-volume-correction'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Statistical correction of statistics-volume pipeline:\n'
                             'http://clinica.run/doc/Pipelines/Statistics_Volume/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES

        # Clinica compulsory arguments
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("t_map",
                                  help='t-statistics map')
        clinica_comp.add_argument("height_threshold", type=float,
                                  help='T value corresponding to an uncorrected p-value of 0.001')
        clinica_comp.add_argument("FWEp", type=float,
                                  help='height threshold (i.e. voxel-level (= peak) threshold)')
        clinica_comp.add_argument("FDRp", type=float,
                                  help='height threshold (i.e. voxel-level (= peak) threshold)')
        clinica_comp.add_argument("FWEc", type=int,
                                  help='extent threshold (i.e. cluster size threshold)')
        clinica_comp.add_argument("FDRc", type=int,
                                  help='extent threshold (i.e. cluster size threshold)')

        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-nc", "--n_cuts", default=8, type=int,
                              help='Number of cuts along each direction')

        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()

    def run_command(self, args):
        from networkx import Graph
        from .statistics_volume_correction_pipeline import StatisticsVolumeCorrection
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            't_map': args.t_map,
            'height_threshold': args.height_threshold,
            'FWEp': args.FWEp,
            'FDRp': args.FDRp,
            'FWEc': args.FWEp,
            'FDRc': args.FDRp,
            'n_cuts': args.n_cuts
        }

        pipeline = StatisticsVolumeCorrection(
            caps_directory=self.absolute_path(args.caps_directory),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8

"""
This module contains functions to deal with longitudinal datasets. Currently, Clinica Pipeline class and clinica/utils
folder can not handle the case where we need to manipulate longitudinal IDs.

When a new longitudinal pipeline will be developed into Clinica, refactoring will be needed.
"""


def extract_subject_session_longitudinal_ids_from_filename(bids_or_caps_files):
    """Extract participant/session/longitudinal IDs
    (e.g. ['sub-CLNC01', 'sub-CLNC01']/['ses-M00', 'ses-M18']/['long-M00M18', 'long-M00M18'])
    from `bids_or_caps_files`.

    TODO: Find a way to merge with utils/filemanip.py::extract_subjects_sessions_from_filename into one util
    """
    import re
    id_bids_or_caps_files = [re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)_(long-[a-zA-Z0-9]+)', file).group()
                             for file in bids_or_caps_files]
    split = [image_id.split('_')
             for image_id in id_bids_or_caps_files]
    part_ids = [p_id[0] for p_id in split]
    sess_ids = [s_id[1] for s_id in split]
    long_ids = [l_id[2] for l_id in split]
    return part_ids, sess_ids, long_ids


def read_part_sess_long_ids_from_tsv(tsv_file):
    """Extract participant, session and longitudinal from TSV file.

    TODO: Find a way to merge with utils/filemanip.py::read_participant_tsv into one util
    """
    import os
    import pandas
    from colorama import Fore
    from clinica.utils.exceptions import ClinicaException

    if not os.path.isfile(tsv_file):
        raise ClinicaException(
            "\n%s[Error] The TSV file you gave is not a file.%s\n"
            "\n%sError explanations:%s\n"
            " - Clinica expected the following path to be a file: %s%s%s\n"
            " - If you gave relative path, did you run Clinica on the good folder?" %
            (Fore.RED, Fore.RESET,
             Fore.YELLOW, Fore.RESET,
             Fore.BLUE, tsv_file, Fore.RESET)
        )
    df = pandas.read_csv(tsv_file, sep='\t')

    def check_key_in_data_frame(file, data_frame, key):
        if key not in list(data_frame.columns.values):
            raise ClinicaException(
                "\n%s[Error] The TSV file does not contain %s column (path: %s)%s" %
                (Fore.RED, key, file, Fore.RESET)
            )
    check_key_in_data_frame(tsv_file, df, 'participant_id')
    check_key_in_data_frame(tsv_file, df, 'session_id')
    check_key_in_data_frame(tsv_file, df, 'long_id')

    participants = list(df.participant_id)
    sessions = list(df.session_id)
    longs = list(df.long_id)

    # Remove potential whitespace in participant, session or longitudinal ID
    return [sub.strip(' ') for sub in participants], \
           [ses.strip(' ') for ses in sessions],\
           [lng.strip(' ') for lng in longs]


def save_part_sess_long_ids_to_tsv(participant_ids, session_ids, long_ids, out_folder, file_name=None):
    """Save participant, session and longitudinal IDs to TSV file.
    TODO: Find a way to merge with utils/save_participants_sessions.py::read_participant_tsv into one util
    """
    import os
    import errno
    import pandas
    from clinica.utils.stream import cprint

    try:
        os.makedirs(out_folder)
    except OSError as e:
        if e.errno != errno.EEXIST:  # EEXIST: folder already exists
            raise e

    if file_name:
        tsv_file = os.path.join(out_folder, file_name)
    else:
        tsv_file = os.path.join(out_folder, 'participants.tsv')

    try:
        data = pandas.DataFrame({
            'participant_id': participant_ids,
            'session_id': session_ids,
            'long_id': long_ids
        })
        data.to_csv(tsv_file, sep='\t', index=False, encoding='utf-8')
    except Exception as e:
        cprint("Impossible to save %s with pandas" % tsv_file)
        raise e


def extract_participant_long_ids_from_filename(caps_files):
    """
    TODO: Find a way to merge with utils/filemanip.py::extract_subjects_sessions_from_filename into one util
    """
    import re
    caps_files = [re.search(r'(sub-[a-zA-Z0-9]+)_(long-[a-zA-Z0-9]+)', file).group()
                  for file in caps_files]
    split = [image_id.split('_')
             for image_id in caps_files]
    part_ids = [p_id[0] for p_id in split]
    long_ids = [l_id[1] for l_id in split]
    return part_ids, long_ids


def grab_image_ids_from_caps_directory(caps_dir):
    """
    Parse CAPS directory to extract participants, sessions and longitudinal IDs.

    Note:
        This function is a simplified version of create_subs_sess_list for CAPS folders with longitudinal IDs

    TODO: Find a way to merge with iotools/utils/data_handling::create_subs_sess_list into one util

    Use case:
    - CAPS
        - subjects
            - sub-CLNC01
                - long-M00M18
                - ses-M00
                - ses-M18
                - ses-M36
            - sub-CLNC02
                - ses-M00
                - ses-M18

    part_ids = ["sub-CLNC01",  "sub-CLNC01",  "sub-CLNC01" ]
    sess_ids = ["ses-M00",     "ses-M18",     "ses-M36"    ]
    long_ids = ["long-M00M18", "long-M00M18", "long-M00M18"]
    (sub-CLNC02 does not have longitudinal ID so it does not appear on the result)

    Args:
        caps_dir (str): Path to the CAPS directory.
    """
    from glob import glob
    import os

    participants_paths = glob(os.path.join(caps_dir, 'subjects', '*sub-*'))

    participants_paths.sort()
    if len(participants_paths) == 0:
        raise IOError('Dataset empty or not CAPS compliant.')

    part_ids = []
    sess_ids = []
    long_ids = []
    for part_path in participants_paths:
        part_id = part_path.split(os.sep)[-1]
        session_paths = glob(os.path.join(part_path, '*ses-*'))
        longitudinal_paths = glob(os.path.join(part_path, '*long-*'))
        if len(session_paths) and len(longitudinal_paths):
            for ses_path in session_paths:
                sess_id = ses_path.split(os.sep)[-1]
                for long_path in longitudinal_paths:
                    long_id = long_path.split(os.sep)[-1]

                    part_ids.append(part_id)
                    sess_ids.append(sess_id)
                    long_ids.append(long_id)

    return part_ids, sess_ids, long_ids



# coding: utf8

"""
Utils for the implementation of t1-volume-tissue-segmentation pipeline
"""

import os
import numpy as np
from nipype.interfaces.spm.base import SPMCommandInputSpec, SPMCommand
from nipype.interfaces.base import TraitedSpec, OutputMultiPath, InputMultiPath, File, traits
from nipype.utils.filemanip import filename_to_list, list_to_filename


def t1w_container_from_filename(t1w_filename):
    """
    Extracts <participant_id> & <sesssion_id> from BIDS <t1w_filename> and
    returns CAPS path.
    """
    import re
    from os.path import join
    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)_', t1w_filename)

    if m is None:
        raise ValueError('Input filename is not in a BIDS or CAPS compliant format. It does not contain the subject' +
                         ' and session information.')

    participant_id = m.group(1)
    session_id = m.group(2)

    return join('subjects', participant_id, session_id, 't1', 'spm', 'segmentation')


def init_input_node(t1w):
    """
    Extracts "sub-<participant_id>_ses-<session_label>" from input node
    and prints begin message.
    """
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    subject_id = get_subject_id(t1w)
    print_begin_image(subject_id)

    return subject_id, t1w


def print_end_pipeline(subject_id, final_file):
    """
    Display end message for <subject_id> when <final_file> is connected.
    """
    from clinica.utils.ux import print_end_image
    print_end_image(subject_id)


def zip_list_files(class_images, zip_files=False):
    from clinica.utils.filemanip import zip_nii

    if zip_files:
        return [zip_nii(tissue, True) for tissue in class_images]

    return [tissue for tissue in class_images]


def get_tissue_tuples(tissue_map, tissue_classes, dartel_tissues, save_warped_unmodulated, save_warped_modulated):
    """
    Method to obtain the list of tuples, one for each tissue class, with the following fields:
     - tissue probability map (4D), 1-based index to frame
     - number of gaussians
     - which maps to save [Native, DARTEL] - a tuple of two boolean values
     - which maps to save [Unmodulated, Modulated] - a tuple of two boolean values

    Args:
        tissue_map: Path to tissue maps
        tissue_classes: Classes of images to obtain from segmentation. Ex: [1,2,3] is GM, WM and CSF
        dartel_tissues: Classes of images to save for DARTEL template calculation. Ex: [1] is only GM'
        save_warped_unmodulated: Save warped unmodulated images for tissues specified in --tissue_classes
        save_warped_modulated: Save warped modulated images for tissues specified in --tissue_classes

    Returns:
        List of tuples according to NewSegment input por tissues
    """
    tissues = []

    for i in range(1, 7):
        n_gaussians = 2

        if i == 4 or i == 5:
            n_gaussians = i - 1

        native_space = False
        dartel_input = False
        warped_unmodulated = False
        warped_modulated = False

        if i in tissue_classes:
            native_space = True
            if save_warped_unmodulated:
                warped_unmodulated = True
            if save_warped_modulated:
                warped_modulated = True

        if i in dartel_tissues:
            dartel_input = True

        tissues.append(((tissue_map, i),
                        n_gaussians,
                        (native_space, dartel_input),
                        (warped_unmodulated, warped_modulated)))
    return tissues


class ApplySegmentationDeformationInput(SPMCommandInputSpec):

    deformation_field = File(
        exists=True,  mandatory=True,
        field='comp{1}.def',
        desc='SPM Segmentation deformation file')
    in_files = InputMultiPath(
        File(exists=True), mandatory=True,
        field='out{1}.pull.fnames',
        desc='Files on which deformation field is applied')
    interpolation = traits.Range(
        low=0, high=7,
        field='out{1}.pull.interp',
        desc='degree of b-spline used for interpolation')
    mask = traits.Int(
        0,
        usedefault=True,
        field='out{1}.pull.mask',
        desc='image masking')
    fwhm = traits.List(
        traits.Float(0),
        field='out{1}.pull.fwhm',
        minlen=3, maxlen=3,
        desc='3-element list (opt)')


class ApplySegmentationDeformationOutput(TraitedSpec):
    out_files = OutputMultiPath(File(exists=True),
                                desc='Transformed files')


class ApplySegmentationDeformation(SPMCommand):
    """ Uses SPM to apply a deformation field obtained from Segmentation routine to a given file

    Examples
    --------

    >>> import clinica.pipelines.t1_volume_tissue_segmentation.t1_volume_tissue_segmentation_utils as seg_utils
    >>> inv = seg_utils.ApplySegmentationDeformation()
    >>> inv.inputs.in_files = 'T1w.nii'
    >>> inv.inputs.deformation = 'y_T1w.nii'
    >>> inv.run() # doctest: +SKIP
    """

    input_spec = ApplySegmentationDeformationInput
    output_spec = ApplySegmentationDeformationOutput

    _jobtype = 'util'
    _jobname = 'defs'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt == 'deformation_field':
            return np.array([list_to_filename(val)], dtype=object)
        if opt == 'in_files':
            return np.array(filename_to_list(val), dtype=object)
        return val

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_files'] = []
        for filename in self.inputs.in_files:
            _, fname = os.path.split(filename)
            outputs['out_files'].append(os.path.realpath('w%s' % fname))
        return outputs



# coding: utf8

import clinica.engine as ce


class T1FreeSurferVisualizer(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 't1-freesurfer'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Cross-sectional pre-processing of T1w images with FreeSurfer:\n'
                             'http://clinica.run/doc/Pipelines/T1_FreeSurfer/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES

        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("participant_id",
                                  help='Participant ID (e.g. sub-CLNC01).')
        clinica_comp.add_argument("session_id",
                                  help='Session ID (e.g. ses-M00).')

    def run_command(self, args):
        """
        """
        import os
        from clinica.utils.stream import cprint
        from clinica.utils.check_dependency import check_freesurfer
        import subprocess

        check_freesurfer()

        participant_id = args.participant_id
        session_id = args.session_id
        subject_id = participant_id + '_' + session_id

        cprint('Visualizing outputs of t1-freesurfer pipeline for '
               + participant_id + ' at session ' + session_id)
        caps_participant = os.path.join(
            args.caps_directory, 'subjects', participant_id, session_id,
            't1', 'freesurfer_cross_sectional', subject_id
        )

        # Check that
        # 1) subject and session are present in CAPS
        # 2) a FreeSurfer folder has been launched
        if not os.path.exists(caps_participant):
            if os.path.exists(os.path.join(args.caps_directory,
                                           'subjects',
                                           participant_id,
                                           session_id)):
                raise RuntimeError('Subject ' + participant_id + ' and session '
                                   + session_id + ' exists in CAPS but no '
                                   + 'FreeSurfer related directory found')
            else:
                raise RuntimeError('Subject ' + participant_id + ' and session '
                                   + session_id + ' are not present in CAPS '
                                   + args.caps_directory)

        # Check that freesurfer has run properly
        log_file = os.path.join(caps_participant,
                                'scripts',
                                'recon-all-status.log')
        if os.path.isfile(log_file):
            last_line = str(subprocess.check_output(['tail', '-1', log_file]))
            if 'finished without error' not in last_line.lower():
                raise ValueError('FreeSurfer did not mark subject '
                                 + subject_id + ' as -finished without '
                                 + 'error-. If the pipeline is still running'
                                 + ', wait for it to finish. Otherwise, '
                                 + 'relaunch it.')
        else:
            raise FileNotFoundError('File ' + log_file + ' was not found. '
                                    + 't1-freesurfer must be run before '
                                    + 'using: clinica visualize t1-freesurfer')

        # Check that files exists
        files = {'nu': os.path.join(caps_participant, 'mri', 'nu.mgz'),
                 'aseg': os.path.join(caps_participant, 'mri', 'aseg.mgz'),
                 'lh_white': os.path.join(caps_participant, 'surf', 'lh.white'),
                 'lh_pial': os.path.join(caps_participant, 'surf', 'lh.pial'),
                 'rh_white': os.path.join(caps_participant, 'surf', 'rh.white'),
                 'rh_pial': os.path.join(caps_participant, 'surf', 'rh.pial')}

        files_not_found = []
        for element in files:
            if not os.path.exists(files[element]):
                files_not_found.append(files[element])
        if len(files_not_found) > 0:
            error_string = 'The following file(s) has(ve) not been found :\n'
            for f in files_not_found:
                error_string = error_string + f + '\n'
            raise FileNotFoundError(error_string)

        # Launch command
        command_line = \
            'freeview' \
            ' -v' \
            ' %s:name=T1w-nu-corrected' \
            ' %s:colormap=lut:opacity=0.2:name=Segmentation' \
            ' -f' \
            ' %s:edgecolor=blue:name=Left_WM/GM_Interface' \
            ' %s:edgecolor=green:name=Left_GM/CSF_Interface' \
            ' %s:edgecolor=blue:name=Right_WM/GM_Interface' \
            ' %s:edgecolor=green:name=Right_GM/CSF_Interface' % \
            (files['nu'],
             files['aseg'],
             files['lh_white'],
             files['lh_pial'],
             files['rh_white'],
             files['rh_pial'])
        cprint('Running the following command: %s' % command_line)
        os.system(command_line)



# coding: utf8

"""
Convert the NIFD dataset into BIDS.
"""


def convert_images(path_to_dataset, bids_dir, path_to_clinical):
    # Conversion of the entire dataset in BIDS
    """Scans available files in the path_to_dataset,
    identifies the patients that have images described by the json file,
    converts the image with the highest quality for each category"""

    import os
    from clinica.iotools.converters.nifd_to_bids.utils.conv_image_folders import get_all_med_name, dict_conversion, \
        get_descriptors
    from clinica.iotools.converters.nifd_to_bids.nifd_utils import get_patients_source_files, \
        filter_patients_source_files, \
        collect_conversion_tuples, convert
    from clinica.iotools.converters.nifd_to_bids.utils.manage_conflicts import Manage_conflicts
    from clinica.iotools.converters.nifd_to_bids.utils.patient import Patient
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.nifd_to_bids.preprocessing.parse_ida import process_ida
    from clinica.iotools.converters.nifd_to_bids.preprocessing.update_clinical_info import update_info_clinical

    path_converter = os.path.join(os.path.dirname(os.path.realpath(__file__)))
    path_conflicts = os.path.join(path_converter, 'config_files', 'unique_conflicts.txt')
    path_to_clinical_file = os.path.join(path_to_clinical, 'NIFD_Clinical_Data_2017_final_updated.xlsx')

    path_to_ida = os.path.join(path_to_clinical, 'ida.tsv')

    path_to_clinical_info = os.path.join(path_to_clinical, 'clinical_info.tsv')

    name_ida = None
    for filename in os.listdir(path_to_clinical):
        if filename.lower().startswith('idasearch') and filename.lower().endswith('.csv'):
            name_ida = filename
    if name_ida is None:
        name_ida = 'idaSearch_all.csv'

    path_idaSearch = os.path.join(path_to_clinical, name_ida)
    path_DataDictionary_NIFD_2017 = os.path.join(path_to_clinical, 'DataDictionary_NIFD_2017.10.18.xlsx')

    # Pre-processing step, to be executed the first time the converter is used.
    if not os.path.isfile(path_to_ida):
        if os.path.isfile(path_idaSearch):
            cprint(
                "ida.tsv was not found in the clinical data directory, "
                "ida.tsv will be created from idaSearch_<date_of_download>_all.csv")

        else:
            cprint(
                "\nida.tsv does not exist and idaSearch_<date_of_download>_all.csv "
                "was not found in the clinical data directory,"
                " to create it please enter path/to/idaSearch_all.csv :")
            path_idaSearch = input()
            path_idaSearch = path_idaSearch.strip(' ')
        cprint("Creating ida.tsv ...")
        process_ida(path_idaSearch, path_to_clinical)
        assert os.path.isfile(path_to_ida), 'Failed to create ida.tsv'
        cprint("ida.tsv successfully created !")

    if not os.path.isfile(path_to_clinical_info):
        if os.path.isfile(path_idaSearch):
            cprint(
                "clinical_info.tsv was not found in the clinical data directory, "
                "clinical_info.tsv will be created from DataDictionary_NIFD_2017.10.18.xlsx")

        else:
            cprint(
                "\nclinical_info.tsv does not exist and DataDictionary_NIFD_2017.10.18.xlsx was not found in the "
                "clinical data directory"
                ", to create it please enter path/to/DataDictionary_NIFD_2017.10.18.xlsx :")
            path_DataDictionary_NIFD_2017 = input()
            path_DataDictionary_NIFD_2017 = path_DataDictionary_NIFD_2017.strip(' ')
        cprint("Creating clinica_info.tsv ...")

        path_clinicals = os.path.join(path_converter, '..', 'clinical_data_bids_correspondence')

        update_info_clinical(path_DataDictionary_NIFD_2017, path_clinicals, path_to_clinical_file, path_to_clinical)
        assert os.path.isfile(path_to_clinical_info), 'Failed to create clinical_info.tsv'
        cprint("clinical_info.tsv successfully created !")

    cprint("Parsing files to be converted...")
    medical_images = get_all_med_name(path_to_dataset)
    assert len(medical_images) > 0, 'The dataset contains no medical image'

    descriptors = get_descriptors(os.path.join(path_converter, 'config_files'))
    assert len(descriptors) > 0, 'Failed to load the descriptors'

    # equivalences['medical_image_name'] = (Descriptor_instance, modalityLabel)
    # ex : equivalences['T1_mprage_S3_DIS3D'] -> (<Descriptor.Descriptor object at 0x105459ac8>, 'T1w')
    equivalences = dict_conversion(medical_images, descriptors)

    # patients_source_files[subject_ID] = [paths_to_all_medical_images_of_subject]
    # Only contains files that are available in the provided dataset
    patients_source_files = get_patients_source_files(path_to_dataset, path_to_ida)
    patients_source_files = filter_patients_source_files(patients_source_files, path_to_dataset, descriptors)
    patients_source_files = {pat: patients_source_files[pat] for pat in patients_source_files if
                             patients_source_files[pat] != []}

    cf = Manage_conflicts(path_conflicts)

    to_convert = []
    for pat in patients_source_files:
        path_patient = os.path.join(path_to_dataset, pat)

        patient = Patient(pat, path_patient, path_to_ida)
        final_bids = patient.clean_conflicts(equivalences, descriptors, patients_source_files[pat], cf, pat)

        to_convert.extend(collect_conversion_tuples(final_bids, bids_dir, patient))

    assert to_convert != [], "No Dicom files to convert!"
    cprint("Converting files to Nifti")

    # Converting only images that have not been already converted,
    # the converter does not have to restart from scratch if something fails
    final_convert = []
    for tuple in to_convert:
        if not os.path.isfile(tuple[1] + '.nii.gz'):
            final_convert.append(tuple)

    convert(final_convert)
    return to_convert


def convert_clinical_data(bids_dir, path_to_clinical, to_convert):
    # clinical specifications in BIDS
    import os
    from clinica.iotools.converters.nifd_to_bids.utils.parse_clinical import Parse_clinical
    from clinica.utils.stream import cprint

    path_to_ida = os.path.join(path_to_clinical, 'ida.tsv')
    assert os.path.isfile(path_to_ida), 'Failed to create ida.tsv'

    cprint('Creating clinical data files')

    pc = Parse_clinical(path_to_clinical)
    pc.make_all(bids_dir)
    pc.make_all_scans(to_convert)



# coding: utf-8

"""
Module for converting T1 of ADNI
"""


def convert_adni_t1(source_dir, csv_dir, dest_dir, subjs_list=None):
    """Convert T1 MR images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    from os import path

    from pandas.io import parsers
    from colorama import Fore

    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = parsers.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of T1 images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_t1_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of T1 images found. Exporting images into BIDS ...')
    paths_to_bids(images, dest_dir, 't1')
    cprint(Fore.GREEN + 'T1 conversion done.' + Fore.RESET)


def compute_t1_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """Compute the paths to T1 MR images and store them in a tsv file

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: a dataframe with all the paths to the T1 MR images that will be converted into BIDS

    """

    from os import path, makedirs

    import pandas as pd

    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import visits_to_timepoints, find_image_path

    t1_col_df = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date',
                 'Study_ID', 'Field_Strength', 'Series_ID', 'Image_ID', 'Original']
    t1_df = pd.DataFrame(columns=t1_col_df)
    t1_dfs_list = []

    # Loading needed .csv files
    adni_merge = pd.read_csv(path.join(csv_dir, 'ADNIMERGE.csv'), sep=',', low_memory=False)
    mprage_meta = pd.read_csv(path.join(csv_dir, 'MPRAGEMETA.csv'), sep=',', low_memory=False)
    mri_quality = pd.read_csv(path.join(csv_dir, 'MRIQUALITY.csv'), sep=',', low_memory=False)
    mayo_mri_qc = pd.read_csv(path.join(csv_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv'), sep=',', low_memory=False)
    # Keep only T1 scans
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == 'T1']

    # We will convert the images for each subject in the subject list
    for subj in subjs_list:

        # Filter ADNIMERGE, MPRAGE METADATA and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values('EXAMDATE')

        mprage_meta_subj = mprage_meta[mprage_meta.SubjectID == subj]
        mprage_meta_subj = mprage_meta_subj.sort_values('ScanDate')

        mri_quality_subj = mri_quality[mri_quality.RID == int(subj[-4:])]
        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(subj, mprage_meta_subj, adnimerge_subj, "T1", "Visit", "ScanDate")

        for visit_info in visits.keys():
            cohort = visit_info[1]
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            image_dict = None

            if cohort in ('ADNI1', 'ADNIGO', 'ADNI2'):
                image_dict = adni1go2_image(subj, timepoint, visit_str, mprage_meta_subj, mri_quality_subj,
                                            mayo_mri_qc_subj,
                                            preferred_field_strength=1.5 if cohort == 'ADNI1' else 3.0)
            elif cohort == 'ADNI3':
                image_dict = adni3_image(subj, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj)
            else:
                cprint("Subject %s visit %s belongs to an unknown cohort: %s" % (subj, visit_str, cohort))

            if image_dict is not None:
                row_to_append = pd.DataFrame(image_dict, index=['i', ])
                t1_dfs_list.append(row_to_append)

    if t1_dfs_list:
        t1_df = pd.concat(t1_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # Eq_1
                         ('031_S_0830', 'm48'),
                         ('100_S_0995', 'm18'),
                         ('031_S_0867', 'm48'),
                         ('100_S_0892', 'm18'),
                         # Empty folders
                         # ('029_S_0845', 'm24'),
                         # ('094_S_1267', 'm24'),
                         # ('029_S_0843', 'm24'),
                         # ('027_S_0307', 'm48'),
                         # ('057_S_1269', 'm24'),
                         # ('036_S_4899', 'm03'),
                         # ('033_S_1016', 'm120'),
                         # ('130_S_4984', 'm12'),
                         # ('027_S_4802', 'm06'),
                         # ('131_S_0409', 'bl'),
                         # ('082_S_4224', 'm24'),
                         # ('006_S_4960', 'bl'),
                         # ('006_S_4960', 'm03'),
                         # ('006_S_4960', 'm06'),
                         # ('006_S_4960', 'm12'),
                         # ('006_S_4960', 'm24'),
                         # ('006_S_4960', 'm36'),
                         # ('006_S_4960', 'm72'),
                         # ('022_S_5004', 'bl'),
                         # ('022_S_5004', 'm03'),
                         # Several images: T1wa ...
                         ('006_S_4485', 'm84'),
                         ('029_S_2395', 'm72'),
                         ('114_S_6039', 'bl'),
                         ('016_S_4952', 'm48')]

    # Removing known exceptions from images to convert
    if not t1_df.empty:
        error_indices = t1_df.index[t1_df.apply(lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1)]
        t1_df.drop(error_indices, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(t1_df, source_dir, 'T1', 'S', 'Series_ID')

    # Store the paths inside a file called conversion_info inside the input directory
    t1_tsv_path = path.join(dest_dir, 'conversion_info')
    if not path.exists(t1_tsv_path):
        makedirs(t1_tsv_path)
    images.to_csv(path.join(t1_tsv_path, 't1_paths.tsv'), sep='\t', index=False)

    return images


def adni1go2_image(subject_id, timepoint, visit_str, mprage_meta_subj, mri_quality_subj, mayo_mri_qc_subj,
                   preferred_field_strength=3.0):
    """Selects the preferred scan for a subject in a visit, given the subject belongs to ADNI 1, Go or 2 cohorts

    Args:

        subject_id: string containing subject ID
        timepoint: string of visit code in months
        visit_str: string of visit name
        mprage_meta_subj: DataFrame of MPRAGE metadata of images corresponding to the subject
        mri_quality_subj: DatFrame of MR image quality of images corresponding to the subject
        mayo_mri_qc_subj: DatFrame of MAYO Clinic MR image quality of images corresponding to the subject
        preferred_field_strength: Field strength that is preferred in case there are several image acquisitions

    Returns:
        Dictionary containing selected scan information

    """

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars

    filtered_mprage = preferred_processed_scan(mprage_meta_subj, visit_str)

    # If no N3 processed image found (it means there are no processed images at all), get best original image
    if filtered_mprage.empty:
        return original_image(subject_id, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj,
                              preferred_field_strength)

    # If there are images with different magnetic field strength, prefer 1.5T images for ADNI1, 3.0T otherwise
    if len(filtered_mprage.MagStrength.unique()) > 1:
        filtered_mprage = filtered_mprage[filtered_mprage.MagStrength == preferred_field_strength]

    # Sort by Series ID in case there are several images, so we keep the one acquired first
    filtered_mprage = filtered_mprage.sort_values('SeriesID')
    scan = filtered_mprage.iloc[0]

    # Check if selected scan passes QC (if QC exists)
    if not check_qc(scan, subject_id, visit_str, mri_quality_subj):

        # If not passed, look for another scan in the visit from a different acquisition
        filtered_mprage = preferred_processed_scan(mprage_meta_subj, visit_str, unwanted_series_id=[scan.SeriesID])

        if filtered_mprage.empty:
            return None

        scan = filtered_mprage.iloc[0]
        # Check QC for second scan
        if not check_qc(scan, subject_id, visit_str, mri_quality_subj):
            return None

    n3 = scan.Sequence.find('N3')
    # Sequence ends in 'N3' or in 'N3m'
    sequence = scan.Sequence[:n3 + 2 + int(scan.Sequence[n3 + 2] == 'm')]
    sequence = replace_sequence_chars(sequence)

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': False}


def preferred_processed_scan(mprage_meta_subj, visit_str, unwanted_series_id=()):
    """
    Filters content of Dataframe containing MPRAGE metadata to obtain last processed image for the specified visit

    Args:
        mprage_meta_subj: DataFrame of MPRAGE metadata of images corresponding to the subject
        visit_str: Visit name
        unwanted_series_id: Series ID of scans not passing QC

    Returns:
        Dataframe

    """

    mprage_meta_subj = mprage_meta_subj[~mprage_meta_subj.SeriesID.isin(unwanted_series_id)]

    # Get the preferred scan (image series that has been Scaled)
    filtered_mprage = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Processed')
                                       & (mprage_meta_subj.Visit == visit_str)
                                       & mprage_meta_subj.Sequence.str.endswith('Scaled', na=False)]

    # If no preferred image found, get N3 processed image (N3m)
    if filtered_mprage.empty:
        filtered_mprage = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Processed')
                                           & (mprage_meta_subj.Visit == visit_str)
                                           & mprage_meta_subj.Sequence.str.endswith('N3m', na=False)]

    return filtered_mprage


def adni3_image(subject_id, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj):
    """Selects the preferred scan for a subject in a visit, given the subject belongs to ADNI 3 cohort

    Args:
        subject_id: string containing subject ID
        timepoint: string of visit code in months
        visit_str: string of visit name
        mprage_meta_subj: DataFrame of MPRAGE metadata of images corresponding to the subject
        mayo_mri_qc_subj: DatFrame of MAYO Clinic MR image quality of images corresponding to the subject

    Returns:
        Dictionary containing selected scan information

    """

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint

    filtered_scan = mprage_meta_subj[(mprage_meta_subj['Orig/Proc'] == 'Original')
                                     & (mprage_meta_subj.Visit == visit_str)
                                     & mprage_meta_subj.Sequence.str.contains('accel', case=False, na=False)
                                     & ~mprage_meta_subj.Sequence.str.lower().str.endswith('_nd', na=False)]

    if filtered_scan.empty:
        cprint('NO MPRAGE Meta for ADNI3: ' + subject_id + ' for visit ' + timepoint + ' - ' + visit_str)
        return None

    scan = select_scan_from_qc(filtered_scan, mayo_mri_qc_subj, preferred_field_strength=3.0)
    sequence = replace_sequence_chars(scan.Sequence)

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': True}


def original_image(subject_id, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj, preferred_field_strength=3.0):
    """Selects the preferred scan for a subject in a visit, given the scan is not preprocessed

    Args:
        subject_id:string containing subject ID
        timepoint: string of visit code in months
        visit_str: string of visit name
        mprage_meta_subj: DataFrame of MPRAGE metadata of images corresponding to the subject
        mayo_mri_qc_subj: DatFrame of MAYO Clinic MR image quality of images corresponding to the subject
        preferred_field_strength: Field strength that is preferred in case there are several image acquisitions


    Returns:
        Dictionary containing selected scan information

    """

    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars
    from clinica.utils.stream import cprint

    mprage_meta_subj_orig = mprage_meta_subj[mprage_meta_subj['Orig/Proc'] == 'Original']

    cond_mprage = ((mprage_meta_subj_orig.Visit == visit_str)
                   & mprage_meta_subj_orig.Sequence.str.contains("mprage|mp-rage|mp rage", case=False, na=False)
                   & ~mprage_meta_subj_orig.Sequence.str.contains("2", na=False))

    cond_spgr = ((mprage_meta_subj_orig.Visit == visit_str)
                 & mprage_meta_subj_orig.Sequence.str.contains("spgr", case=False, na=False)
                 & ~mprage_meta_subj_orig.Sequence.str.contains("acc", case=False, na=False))

    filtered_scan = mprage_meta_subj_orig[cond_mprage | cond_spgr]

    if filtered_scan.empty:
        cprint('NO MPRAGE Meta: ' + subject_id + ' for visit ' + timepoint + ' - ' + visit_str)
        return None

    scan = select_scan_from_qc(filtered_scan, mayo_mri_qc_subj, preferred_field_strength)
    sequence = replace_sequence_chars(scan.Sequence)

    return {'Subject_ID': subject_id,
            'VISCODE': timepoint,
            'Visit': visit_str,
            'Sequence': sequence,
            'Scan_Date': scan.ScanDate,
            'Study_ID': str(scan.StudyID),
            'Series_ID': str(scan.SeriesID),
            'Image_ID': str(scan.ImageUID),
            'Field_Strength': scan.MagStrength,
            'Original': True}


def select_scan_from_qc(scans_meta, mayo_mri_qc_subj, preferred_field_strength):
    """Selects a scan from a list of scans taking into account available QC

    Args:
        scans_meta: DataFrame containing the metadata for images to choose among
        mayo_mri_qc_subj: DatFrame of MAYO Clinic MR image quality of images corresponding to the subject
        preferred_field_strength: Field strength that is preferred in case there are several image acquisitions

    Returns:
        DataFrame row containing selected scan

    """
    import numpy as np

    multiple_mag_strength = False

    # Select preferred_field_strength images
    if len(scans_meta.MagStrength.unique()) > 1:
        multiple_mag_strength = True
        # Save for later the scans with the non preferred magnetic field strength
        not_preferred_scan = scans_meta[scans_meta.MagStrength != preferred_field_strength]
        # Filtering to keep only the scans with the preferred magnetic field strength
        scans_meta = scans_meta[scans_meta.MagStrength == preferred_field_strength]

    if scans_meta.MagStrength.unique()[0] == 3.0:

        id_list = scans_meta.ImageUID.unique()
        image_ids = ['I' + str(imageuid) for imageuid in id_list]
        int_ids = [int(imageuid) for imageuid in id_list]
        images_qc = mayo_mri_qc_subj[mayo_mri_qc_subj.loni_image.isin(image_ids)]

        if not images_qc.empty:
            selected_image = None
            # We check if there is a single selected series image
            if np.sum(images_qc.series_selected) == 1:
                selected_image = images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
            # Otherwise we select the one with the best QC if available
            else:
                images_not_rejected = images_qc[images_qc.series_quality < 4]

                if images_not_rejected.empty:

                    # There are no images that passed the qc
                    # so we'll try to see if there are other images without qc,
                    # otherwise return None
                    qc_ids = set([int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()])
                    no_qc_ids = list(set(int_ids) - qc_ids)

                    # If none of images passed qc the scan is None, otherwise:
                    if no_qc_ids:
                        no_qc_scans_meta = scans_meta[scans_meta.ImageUID.isin(no_qc_ids)]
                        return select_scan_no_qc(no_qc_scans_meta)
                else:
                    # We select the image with the best (lower) QC.
                    # If no positive QC available we choose image with -1 (not performed)
                    series_quality = [q if q > 0 else 4 for q in list(images_not_rejected.series_quality)]
                    best_q = np.amin(series_quality)
                    if best_q == 4:
                        best_q = -1
                    images_best_qc = images_not_rejected[images_not_rejected.series_quality == best_q]
                    if len(images_best_qc) == 1:
                        selected_image = images_best_qc.iloc[0].loni_image[1:]
                    else:
                        best_ids = [int(x[1:]) for x in images_best_qc.loni_image.unique()]
                        best_qc_meta = scans_meta[scans_meta.ImageUID.isin(best_ids)]
                        return select_scan_no_qc(best_qc_meta)

            # If we did not find an image passing QC for the preferred magnetic field strength,
            # then we will choose between the other available images with other magnetic field strength
            if selected_image is None and multiple_mag_strength:
                scans_meta = not_preferred_scan
            else:
                scan = scans_meta[scans_meta.ImageUID == int(selected_image)].iloc[0]
                return scan

    # 1.5T (no available QC)
    return select_scan_no_qc(scans_meta)


def select_scan_no_qc(scans_meta):
    """Selects a scan from available scans in case there is not an available QC

    Args:
        scans_meta: DataFrame containing the metadata for images to choose among

    Returns:
        DataFrame row containing selected scan

    """

    # We choose the first scan (not containing repeat in name)
    selected_scan = scans_meta[~scans_meta.Sequence.str.contains("repeat", case=False, na=False)]

    if selected_scan.empty:
        selected_scan = scans_meta

    scan = selected_scan.iloc[0]
    return scan


def check_qc(scan, subject_id, visit_str, mri_quality_subj):
    """Checks if a scan has passed check quality, if it was performed

    Args:
        scan: DataFrame row containing scan information
        subject_id: string containing subject ID
        visit_str: string of visit name
        mri_quality_subj: DatFrame of MR image quality of images corresponding to the subject

    Returns:
        boolean, True if image passed QC or if there is not available QC, False elsewhere

    """
    from clinica.utils.stream import cprint

    # Check if QC exists for image series
    qc = mri_quality_subj[mri_quality_subj.LONIUID == 'S' + str(scan.SeriesID)]

    # If QC exists and failed we keep the other scan (in case 2 scans were performed)
    if not qc.empty and qc.iloc[0].PASS != 1:

        cprint('QC found but NOT passed')
        cprint('Subject ' + subject_id + ' for visit ' + visit_str +
               ' - Series: ' + str(scan.SeriesID) + ' - Study: ' + str(scan.StudyID))
        return False
    return True



# coding: utf-8

"""
Module for converting FDG PET of ADNI
"""


def convert_adni_fdg_pet(source_dir, csv_dir, dest_dir, subjs_list=None):
    """Convert FDG PET images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list
    """
    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of FDG PET images. Output will be stored in %s.' % path.join(dest_dir, 'conversion_info'))
    images = compute_fdg_pet_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of FDG PET images found. Exporting images into BIDS ...')
    paths_to_bids(images, dest_dir, 'fdg')
    cprint(Fore.GREEN + 'FDG PET conversion done.' + Fore.RESET)


def compute_fdg_pet_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """Compute the paths to the FDG PET images and store them in a tsv file

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: a dataframe with all the paths to the PET images that will be converted into BIDS

    """

    import pandas as pd
    import os
    from os import path
    from clinica.iotools.converters.adni_to_bids.adni_utils import get_images_pet, find_image_path

    pet_fdg_col = ['Phase', 'Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date', 'Study_ID',
                   'Series_ID', 'Image_ID', 'Original']
    pet_fdg_df = pd.DataFrame(columns=pet_fdg_col)
    pet_fdg_dfs_list = []

    # Loading needed .csv files
    petqc = pd.read_csv(path.join(csv_dir, 'PETQC.csv'), sep=',', low_memory=False)
    petqc3 = pd.read_csv(path.join(csv_dir, 'PETC3.csv'), sep=',', low_memory=False)
    pet_meta_list = pd.read_csv(path.join(csv_dir, 'PET_META_LIST.csv'), sep=',', low_memory=False)

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list['Subject'] == subj]

        if subject_pet_meta.empty:
            continue

        # QC for FDG PET images for ADNI 1, GO and 2
        pet_qc_1go2_subj = petqc[(petqc.PASS == 1) & (petqc.RID == int(subj[-4:]))]

        # QC for FDG PET images for ADNI 3
        pet_qc3_subj = petqc3[(petqc3.SCANQLTY == 1) & (petqc3.RID == int(subj[-4:]))]
        pet_qc3_subj.insert(0, 'EXAMDATE', pet_qc3_subj.SCANDATE.to_list())

        # Concatenating visits in both QC files
        pet_qc_subj = pd.concat([pet_qc_1go2_subj, pet_qc3_subj], axis=0, ignore_index=True, sort=False)

        sequences_preprocessing_step = ['Co-registered, Averaged']
        subj_dfs_list = get_images_pet(subj, pet_qc_subj, subject_pet_meta, pet_fdg_col, 'FDG-PET',
                                       sequences_preprocessing_step)
        if subj_dfs_list:
            pet_fdg_dfs_list += subj_dfs_list

    if pet_fdg_dfs_list:
        # Concatenating dataframes into one
        pet_fdg_df = pd.concat(pet_fdg_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # NONAME.nii
                         ('031_S_0294', 'bl'),
                         ('037_S_1421', 'm36'),
                         ('037_S_1078', 'm36'),

                         # Empty folders
                         ('941_S_1195', 'm48'),
                         ('005_S_0223', 'm12')]

    # Removing known exceptions from images to convert
    if not pet_fdg_df.empty:
        error_ind = pet_fdg_df.index[pet_fdg_df.apply(lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors),
                                                      axis=1)]
        pet_fdg_df.drop(error_ind, inplace=True)

    images = find_image_path(pet_fdg_df, source_dir, 'FDG', 'I', 'Image_ID')

    fdg_csv_path = path.join(dest_dir, 'conversion_info')
    if not os.path.exists(fdg_csv_path):
        os.mkdir(fdg_csv_path)
    images.to_csv(path.join(fdg_csv_path, 'fdg_pet_paths.tsv'), sep='\t', index=False)

    return images



# coding: utf8


class Parse_clinical():
    """Creates the clinical files for the BIDS directory"""

    def __init__(self, path_clinical):
        import pandas as pd
        import os.path as path

        self.df_dict_mod = pd.read_csv(path.join(path_clinical, 'clinical_info.tsv'), sep='\t')
        self.df_clinical = pd.read_excel(path.join(path_clinical, 'NIFD_Clinical_Data_2017_final_updated.xlsx'))
        self.df_ida = pd.read_csv(path.join(path_clinical, 'ida.tsv'), sep='\t')

        self.merge_clinical = self.merge_clinical()
        self.df_clinical = self.merge_clinical

    def make_sessions_ida(self, pat_name):
        """
        Preprocesses ida for the left join operated in merge_clinical_scans

        Args:
            pat_name: subject_ID
        """
        def write_ses(num):
            num = str(num)
            num = '0' + num if len(num) == 1 else num
            return 'ses-M' + num

        bloc = self.df_ida[self.df_ida['Subject ID'] == pat_name]
        bloc = bloc[['Visit', 'Study Date', 'Age', 'Weight', 'Research Group']]

        bloc['Visit'] = bloc['Visit'].apply(lambda x: int(x.split(' ')[1]))
        bloc = bloc.sort_values(['Visit'], ascending=[True])
        bloc['Visit'] = bloc['Visit'].apply(lambda x: write_ses(x))
        bloc = bloc.groupby('Visit').first().reset_index()
        bloc.columns = ['session_id', 'examination_date', 'age', 'weight', 'research_group']

        return bloc

    def merge_clinical_scans(self, pat_name):
        """
        Operates a left join between the ida and clinical table
        For a given patient, we have the usual ida file, extended with the information from the clinical table

        Args:
            pat_name: Name of the subject

        Returns:
            dfMerge: pandas dataframe corresponding to the left join
        """
        import pandas as pd
        import warnings

        bloc = self.df_clinical[self.df_clinical['LONI_ID'] == pat_name]

        def parse_date(x):
            x = str(x).split(' ')[0]
            x = x.split('-')
            sol = x[1] + '/' + x[2] + '/' + x[0]
            return sol

        warnings.simplefilter('ignore')
        bloc['CLINICAL_LINKDATE'] = bloc['CLINICAL_LINKDATE'].apply(lambda x: parse_date(x))

        bloc_ida = self.make_sessions_ida(pat_name)
        bloc_ida['examination_date'] = bloc_ida['examination_date'].apply(
            lambda x: '0' + str(x).split(' ')[0] if str(x).split(' ')[0][0] != '0' else str(x).split(' ')[0])

        dfMerge = pd.merge(bloc_ida, bloc, left_on='examination_date', right_on='CLINICAL_LINKDATE')
        dfMerge = dfMerge.drop(columns=['CLINICAL_LINKDATE', 'LONI_ID'])

        return dfMerge

    def merge_clinical(self):
        """
        Operates a left join between the clinical and ida table
        The table that we obtain is the usual clinical table with data, when examination_date and CLINICAL_LINKDATE are the same, from the ida table

        Returns:
            dfSol: pandas dataframe corresponding to the left join
        """
        # Could be optimized
        dfSol = self.df_clinical

        dfSol.insert(4, 'Weight', '')
        dfSol.insert(4, 'Research Group', '')
        dfSol.insert(4, 'Age', '')

        def parse_date(x):
            x = str(x).split(' ')[0]
            x = x.split('-')
            sol = x[1] + '/' + x[2] + '/' + x[0]
            return sol

        curr_sub = None
        for index, row in dfSol.iterrows():
            if row['LONI_ID'] != curr_sub:
                curr_sub = row['LONI_ID']
                dfMerge = self.merge_clinical_scans(curr_sub)
            info = dfMerge[dfMerge['examination_date'] == parse_date(row['CLINICAL_LINKDATE'])]
            if not info.empty:
                dfSol.loc[index, 'Weight'] = float(info['weight'])
                dfSol.loc[index, 'Research Group'] = str(info.iloc[0]['research_group'])
                dfSol.loc[index, 'Age'] = int(info['age'])

        return dfSol

    def get_clinical_ida(self):
        """ Left join clinical data and ida, new version """
        import pandas as pd
        import math

        df_clinical = self.df_clinical.copy()
        df_ida = self.df_ida.copy()
        df_ida['Session_number'] = df_ida['Visit'].apply(lambda x: int(x.split(' ')[-1]))

        df_ida = df_ida.groupby(['Subject ID', 'Study Date']).first().reset_index()
        df_ida = df_ida.sort_values(['Subject ID', 'Session_number']).reset_index()

        def parse_date(x):
            x = x.split('/')
            for i in range(len(x)):
                if len(x[i]) == 1:
                    x[i] = '0' + x[i]
            sol = x[1] + '/' + x[0] + '/' + x[2]
            return sol

        def parse_date2(x):
            x = str(x).split(' ')[0]
            x = x.split('-')
            sol = x[2] + '/' + x[1] + '/' + x[0]
            return sol

        df_ida['Study Date'] = df_ida['Study Date'].apply(lambda x: parse_date(x))
        df_clinical['CLINICAL_LINKDATE'] = df_clinical['CLINICAL_LINKDATE'].apply(lambda x: parse_date2(x))

        dfSol = pd.merge(df_clinical, df_ida, how='left', left_on=['LONI_ID', 'CLINICAL_LINKDATE'],
                         right_on=['Subject ID', 'Study Date'])
        dfSol.insert(0, 'session_id', '')

        dfSol['session_id'] = dfSol['Session_number'].apply(
            lambda x: '' if math.isnan(x) else ('ses-M' + str(int(x)) if x > 10 else 'ses-M0' + str(int(x))))

        return dfSol

    def make_sessions_type(self, pat_name, keep_all=False):
        """ Updated version of make_sessions

        Args:
            pat_name: subject_ID of a patient
            keep_all: if True, include clinical data not linked to a MRI, else include only clinical data linked to a MRI

        Returns:
            bloc: pandas dataframe corresponding to the "sessions.tsv" file"""

        name_clinical, name_BIDS = self.get_names('sessions')
        name_clinical.insert(0, 'session_id')
        name_BIDS.insert(0, 'session_id')
        name_clinical.extend(['Age_x', 'Research Group_x', 'Weight_x'])
        name_BIDS.extend(['age', 'research_group', 'weight'])

        df_clinical_ida = self.get_clinical_ida()
        if not keep_all:
            df_clinical_ida = df_clinical_ida[df_clinical_ida['session_id'] != '']

        bloc = df_clinical_ida[df_clinical_ida['LONI_ID'] == pat_name]
        bloc = bloc[name_clinical]
        bloc.columns = name_BIDS

        return bloc

    def get_names(self, file='sessions'):
        """
        Takes the clinical and BIDS_CLINICA fields that will be included in the clinical file "file"
        Returns them in list form

        Args:
            file: Corresponds to the clinical file that is about to be created, file = "sessions" or "participants"

        Returns:
            name_clinical, name_BIDS: 2 lists of fields
        """
        bloc = self.df_dict_mod[self.df_dict_mod['FILE'] == file]
        name_clinical = bloc['COLUMN_NAME'].tolist()
        name_BIDS = bloc['BIDS_CLINICA'].tolist()

        return name_clinical, name_BIDS

    def make_sessions(self, pat_name):
        """
        Creates the sessions file for a given patient

        Args:
            pat_name: subject_ID of a patient

        Returns:
            bloc: pandas dataframe corresponding to the "sessions.tsv" file
        """
        name_clinical, name_BIDS = self.get_names('sessions')

        bloc = self.df_clinical[self.df_clinical['LONI_ID'] == pat_name]
        bloc = bloc[name_clinical]
        bloc.columns = name_BIDS

        return bloc

    def make_participants(self, pat_list=None):
        """
        Creates the participants file for all patients

        Args:
            pat_list: subject_ID of all patients found in the converted BIDS directory

        Returns:
            bloc: pandas dataframe corresponding to the "participants.tsv" file
        """
        name_clinical, name_BIDS = self.get_names('participants')
        bloc = self.df_clinical.groupby('LONI_ID').first().reset_index()
        bloc = bloc[name_clinical]
        bloc['LONI_ID'] = bloc['LONI_ID'].apply(lambda x: 'sub-NIFD' + x.replace("_", ""))
        bloc.columns = name_BIDS

        if pat_list is not None:
            bloc = bloc[bloc['participant_id'].isin(pat_list)]

        # Enforce the clinica_BIDS convention
        if 'sex' in list(bloc):
            bloc['sex'] = bloc['sex'].apply(lambda x: 'M' if x == 1 else 'F')

        return bloc

    def make_scans(self, path_scans):
        """
        Creates the scans file for a patient's session

        Args:
            path_scans: path to a session for a patient

        Returns:
            bloc: pandas dataframe corresponding to the "participants.tsv" file
        """
        import os

        s = 'filename	scan_id	mri_field\n'
        subs = [f.path.split('/')[-1] for f in os.scandir(path_scans) if f.is_dir()]
        for sub in subs:
            name = os.listdir(os.path.join(path_scans, sub))
            name = [i for i in name if i != '.DS_Store' and (i.endswith('.nii.gz') or i.endswith('.nii'))]

            for n in name:
                s += sub + '/' + n + '\n'
        return s

    def write(self, df, path, name):
        """
        Saves a pandas dataframe

        Args:
            df: a pandas dataframe
            path: Path where the dataframe is to be saved
            name: name of the output file (/!\ do not include the extension, '.tsv' is added in the function)
        """
        import os

        name = os.path.join(path, name) + '.tsv'
        df.to_csv(sep='\t', path_or_buf=name, index=False)

    def make_all(self, pathBIDS):
        """
        Makes the participants.tsv and all sessions.tsv files for all subjects available in the BIDS directory

        Args:
            pathBIDS: path to the BIDS directory
        """
        import os

        pat_list = os.listdir(pathBIDS)
        pat_list = [elt for elt in pat_list if elt.startswith('sub')]

        assert pat_list != [], 'BIDS directory is empty'

        self.write(self.make_participants(pat_list), pathBIDS, 'participants')

        for pat in pat_list:
            path_sessions = os.path.join(pathBIDS, pat)
            pat2 = pat[8] + '_S_' + pat[10:14]

            self.write(self.make_sessions_type(pat2), path_sessions, pat + '_sessions')

    def make_all_scans(self, to_convert):
        """
            Makes the scans.tsv files for all subjects available in the BIDS directory

            Args:
                to_convert: List of tuples of paths (path_in, path_out), computed for the initial image conversion
        """

        import os
        root = '/' + os.path.join(*to_convert[0][1].split('/')[:-4])

        def make_dic_tuples(to_convert, root):
            sol = {}
            dic_pat_sess = {}
            for tuple in to_convert:
                s_path1 = tuple[1].split('/')

                key = os.path.join(root, s_path1[-4], s_path1[-3])

                if s_path1[-4] not in dic_pat_sess:
                    dic_pat_sess[s_path1[-4]] = [s_path1[-3]]
                elif s_path1[-3] not in dic_pat_sess[s_path1[-4]]:
                    dic_pat_sess[s_path1[-4]].append(s_path1[-3])

                if key not in sol:
                    sol[key] = [tuple]
                else:
                    sol[key].append(tuple)

            return sol, dic_pat_sess

        def make_template():
            import pandas as pd

            new_cols = []

            for cell in list(self.df_ida['Imaging Protocol']):
                if isinstance(cell, type('a')):
                    col_to_add = [i.split('=')[0] for i in cell.split(';')]
                    for i in col_to_add:
                        if i not in new_cols:
                            new_cols.append(i)

            new_cols.extend(['Modality', 'Description', 'Type', 'Image ID'])
            new_cols.insert(0, 'filename')
            sol = pd.DataFrame(columns=new_cols)
            return sol

        def extend_line(df_line_ida, template):
            col_values = {}
            s = list(df_line_ida['Imaging Protocol'])[0]
            for coup in s.split(';'):
                col_values[coup.split('=')[0]] = coup.split('=')[1]
            sol = df_line_ida
            for col_name in col_values:
                sol.insert(0, col_name, col_values[col_name])
            for name in list(template):
                if name not in list(sol):
                    sol.insert(0, name, '')
            return sol[list(template)]

        dic_tuples, dic_pat_sess = make_dic_tuples(to_convert, root)
        template = make_template()

        for sub_id in dic_pat_sess:
            for ses_num in dic_pat_sess[sub_id]:
                df_ses = template.copy()
                for tuple in dic_tuples[os.path.join(root, sub_id, ses_num)]:
                    s_path0 = tuple[0].split('/')
                    s_path1 = tuple[1].split('/')
                    filename = os.path.join(s_path1[-2], s_path1[-1]) + '.nii.gz'
                    df_line_ida = self.df_ida[
                        (self.df_ida['Subject ID'] == s_path1[-1][8] + '_S_' + s_path1[-1][10:14]) &
                        (self.df_ida['Visit'] == 'Month ' + str(int(ses_num.split('M')[-1]))) &
                        (self.df_ida['Description'] == s_path0[-3])]

                    if df_line_ida.empty:
                        df_line_ida = self.df_ida[
                            (self.df_ida['Subject ID'] == s_path1[-1][8] + '_S_' + s_path1[-1][10:14]) &
                            (self.df_ida['Visit'] == 'Month ' + str(int(ses_num.split('M')[-1]))) &
                            (self.df_ida['Description'] == s_path0[-3].replace('_', ' '))]

                    # TR_BRAIN_3D_PIB_IR_CTAC -> TR:BRAIN 3D:PIB:IR CTAC
                    if df_line_ida.empty:
                        des = s_path0[-3].split('_')
                        if len(des) == 6:
                            des = des[0] + ':' + des[1] + ' ' + des[2] + ':' + des[3] + ':' + des[4] + ' ' + des[5]
                        else:
                            des = des[0] + ':' + des[1] + ' ' + des[2] + ':' + des[3] + ':' + des[4] + ' ' + des[
                                5] + ' ' + des[6]
                        df_line_ida = self.df_ida[
                            (self.df_ida['Subject ID'] == s_path1[-1][8] + '_S_' + s_path1[-1][10:14]) &
                            (self.df_ida['Visit'] == 'Month ' + str(int(ses_num.split('M')[-1]))) &
                            (self.df_ida['Description'] == des)]

                    df_line_ida.insert(0, 'filename', filename)
                    df_line_ida = extend_line(df_line_ida, template)

                    df_ses = df_ses.append(df_line_ida, ignore_index=True)

                columns = list(df_ses)
                columns[columns.index('Field Strength')] = 'mri_field'
                df_ses.columns = columns

                self.write(df_ses, os.path.join(root, sub_id, ses_num), sub_id + '_' + ses_num + '_scans')



# coding: utf8


import clinica.pipelines.engine as cpe


class T1FreeSurferLongitudinalCorrection(cpe.Pipeline):
    """FreeSurfer Longitudinal correction class

    Returns:
        A clinica pipeline object containing the T1FreeSurferLongitudinalCorrection pipeline
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        pass

    def check_custom_dependencies(self):
        """Check dependencies that cannot be listed in `info.json`.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurferLongitudinalCorrection pipeline is:
                * participant_id (str): Participant ID (e.g. "sub-CLNC01")
                * session_id (str): Session ID associated to `participant_id` (e.g. "ses-M00")
                * long_id (str): Longitudinal ID associated to `participant_id` (e.g. "long-M00" or "long-M00M18")

        Returns:
            A list of (string) input fields name.
        """
        return ['participant_id', 'session_id', 'long_id']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurferLongitudinalCorrection pipeline is:
                * subject_id (str): FreeSurfer ID
                    (e.g. "sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18")

        Returns:
            A list of (string) output fields name.
        """

        return ['subject_id']

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os
        from colorama import Fore

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import T1_FS_DESTRIEUX, T1_FS_T_DESTRIEUX
        from clinica.utils.stream import cprint
        from .longitudinal_utils import (grab_image_ids_from_caps_directory,
                                         read_part_sess_long_ids_from_tsv,
                                         save_part_sess_long_ids_to_tsv,
                                         extract_subject_session_longitudinal_ids_from_filename)
        from .t1_freesurfer_longitudinal_correction_utils import get_processed_images

        #  We re-initialize subjects, sessions and add long IDS since the latter is not handled by Clinica
        if self.tsv_file:
            self.subjects, self.sessions, list_long_id = read_part_sess_long_ids_from_tsv(self.tsv_file)
        else:
            self.subjects, self.sessions, list_long_id = grab_image_ids_from_caps_directory(self.caps_directory)
        # TODO: Find a way to distinguish cross-sectional from longitudinal pipelines in Clinica Core

        # Display image(s) already present in CAPS folder
        # ===============================================
        processed_ids = get_processed_images(self.caps_directory, self.subjects, self.sessions, list_long_id)
        if len(processed_ids) > 0:
            cprint("%sClinica found %s images(s) already processed in CAPS directory:%s" %
                   (Fore.YELLOW, len(processed_ids), Fore.RESET))
            for image_id in processed_ids:
                cprint("%s\t%s%s" % (Fore.YELLOW, image_id.replace('_', ' | '), Fore.RESET))
            if self.overwrite_caps:
                output_folder = "<CAPS>/subjects/<participant_id>/<session_id>/t1/<long_id>/freesurfer_longitudinal/"
                cprint("%s\nOutput folders in %s will be recreated.\n%s" % (Fore.YELLOW, output_folder, Fore.RESET))
            else:
                cprint("%s\nImage(s) will be ignored by Clinica.\n%s" % (Fore.YELLOW, Fore.RESET))
                input_ids = [p_id + '_' + s_id + '_' + l_id for p_id, s_id, l_id in
                             zip(self.subjects, self.sessions, list_long_id)]
                to_process_ids = list(set(input_ids) - set(processed_ids))
                self.subjects, self.sessions, list_long_id = extract_subject_session_longitudinal_ids_from_filename(to_process_ids)

        all_errors = []
        try:
            # Check that t1-freesurfer has run on the CAPS directory
            clinica_file_reader(self.subjects, self.sessions, self.caps_directory, T1_FS_DESTRIEUX)
        except ClinicaException as e:
            all_errors.append(e)

        try:
            # Check that t1-freesurfer-template has run on the CAPS directory
            clinica_file_reader(self.subjects, list_long_id, self.caps_directory, T1_FS_T_DESTRIEUX)
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = 'Clinica faced errors while trying to read files in your CAPS directory.\n'
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaException(error_message)

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        save_part_sess_long_ids_to_tsv(self.subjects, self.sessions, list_long_id, folder_participants_tsv)

        def print_images_to_process(list_participant_id, list_session_id, list_longitudinal_id):
            cprint('The pipeline will be run on the following %s image(s):' % len(list_participant_id))
            for (p_id, s_id, l_id) in zip(list_participant_id, list_session_id, list_longitudinal_id):
                cprint("\t%s | %s | %s" % (p_id, s_id, l_id))

        if len(self.subjects):
            # TODO: Generalize long IDs to the message display
            print_images_to_process(self.subjects, self.sessions, list_long_id)
            cprint('List available in %s' % os.path.join(folder_participants_tsv, 'participants.tsv'))
            cprint('The pipeline will last approximately 10 hours per participant.')

        read_node = npe.Node(name="ReadingFiles",
                             iterables=[
                                 ('participant_id', self.subjects),
                                 ('session_id', self.sessions),
                                 ('long_id', list_long_id),
                             ],
                             synchronize=True,
                             interface=nutil.IdentityInterface(
                                 fields=self.get_input_fields())
                             )
        self.connect([
            (read_node, self.input_node, [('participant_id', 'participant_id')]),
            (read_node, self.input_node, [('session_id', 'session_id')]),
            (read_node, self.input_node, [('long_id', 'long_id')]),
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from .t1_freesurfer_longitudinal_correction_utils import save_to_caps
        import os

        save_to_caps = npe.Node(interface=nutil.Function(
            input_names=['source_dir', 'subject_id', 'caps_dir', 'overwrite_caps'],
            output_names=['image_id'],
            function=save_to_caps),
            name='SaveToCaps')
        save_to_caps.inputs.source_dir = os.path.join(self.base_dir, self.name, 'ReconAll')
        save_to_caps.inputs.caps_dir = self.caps_directory
        save_to_caps.inputs.overwrite_caps = self.overwrite_caps

        self.connect([
            (self.output_node, save_to_caps, [('subject_id', 'subject_id')]),
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import os
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from .t1_freesurfer_longitudinal_correction_utils import (init_input_node,
                                                                  run_recon_all_long,
                                                                  write_tsv_files,
                                                                  move_subjects_dir_to_source_dir)

        # Nodes declaration
        # =================
        # Initialize the pipeline
        init_input = npe.Node(
            interface=nutil.Function(
                input_names=['caps_dir', 'participant_id', 'session_id', 'long_id', 'output_dir'],
                output_names=['subjects_dir'],
                function=init_input_node),
            name='0-InitPipeline')
        init_input.inputs.caps_dir = self.caps_directory
        init_input.inputs.output_dir = os.path.join(self.base_dir, self.name, 'ReconAll')

        # Run recon-all command
        recon_all = npe.Node(
            interface=nutil.Function(
                input_names=['subjects_dir', 'participant_id', 'session_id', 'long_id', 'directive'],
                output_names=['subject_id'],
                function=run_recon_all_long),
            name='1-SegmentationReconAll')
        recon_all.inputs.directive = '-all'

        # Generate TSV files containing a summary of the regional statistics
        create_tsv = npe.Node(interface=nutil.Function(
            input_names=['subjects_dir', 'subject_id'],
            output_names=['subject_id'],
            function=write_tsv_files),
            name='2-CreateTsvFiles')

        # Move $SUBJECT_DIR to source_dir (1 time point case)
        move_subjects_dir = npe.Node(
            interface=nutil.Function(
                input_names=['subjects_dir', 'source_dir', 'subject_id'],
                output_names=['subject_id'],
                function=move_subjects_dir_to_source_dir),
            name='3-MoveSubjectsDir')
        move_subjects_dir.inputs.source_dir = os.path.join(self.base_dir, self.name, 'ReconAll')

        # Connections
        # ===========
        self.connect([
            # Initialize the pipeline
            (self.input_node, init_input, [('participant_id', 'participant_id'),
                                           ('session_id', 'session_id'),
                                           ('long_id', 'long_id')]),
            # Run recon-all command
            (init_input, recon_all, [('subjects_dir', 'subjects_dir')]),
            (self.input_node, recon_all, [('participant_id', 'participant_id'),
                                          ('session_id', 'session_id'),
                                          ('long_id', 'long_id')]),
            # Generate TSV files
            (init_input, create_tsv, [('subjects_dir', 'subjects_dir')]),
            (recon_all,  create_tsv, [('subject_id', 'subject_id')]),
            # Move $SUBJECT_DIR to source_dir (macOS case)
            (init_input, move_subjects_dir, [('subjects_dir', 'subjects_dir')]),
            (create_tsv, move_subjects_dir, [('subject_id', 'subject_id')]),
            # Output node
            (move_subjects_dir, self.output_node, [('subject_id', 'subject_id')]),
        ])



# coding: utf8

"""
Methods to find information in the different pipelines of Clinica
"""


from glob import glob
import pandas as pd
import numpy as np
from os import path
import os
import warnings


def pet_volume_pipeline(caps_dir, df, **kwargs):
    """
    This method merge the data of the PET-Volume pipeline to the merged file
    containing the BIDS information.

    Args:
        caps_dir: the path to the CAPS directory
        df: the DataFrame containing the BIDS information
        atlas_selection: allows to choose the atlas to merge (default = 'all')
        pvc_restriction: gives the restriction on the inclusion or not of the file with the label 'pvc-rbv'
            1       --> only the atlases containing the label will be used
            0       --> the atlases containing the label won't be used
            None    --> all the atlases will be used

    Returns:
         final_df: a DataFrame containing the information of the bids and the pipeline
    """
    from clinica.pipelines.pet_volume.pet_volume_cli import PETVolumeCLI
    instance = PETVolumeCLI()
    pipeline_name = instance.name
    del instance

    # Recovering optional arguments
    if 'pvc_restriction' in kwargs.keys():
        pvc_restriction = kwargs['pvc_restriction']
    else:
        pvc_restriction = None

    if 'atlas_selection' in kwargs.keys():
        atlas_selection = kwargs['atlas_selection']
    else:
        atlas_selection = None

    if 'group_selection' in kwargs.keys():
        group_selection = kwargs['group_selection']
    else:
        group_selection = None

    if pvc_restriction not in [None, 0, 1]:
        raise ValueError('Error in pvc_restriction value: ' + str(pvc_restriction))

    if not path.isdir(caps_dir):
        raise IOError("The CAPS directory doesn't exist")

    pet_path = 'pet' + os.sep + 'preprocessing'
    caps_dir = caps_dir + os.sep + 'subjects'

    # Lists to be sure of the column in case of missing data
    columns = ['pipeline_name', 'group_id', 'atlas_id', 'regions_number', 'first_column_name', 'last_column_name']
    summary_df = pd.DataFrame(columns=columns)
    col_list = []
    group_list = []

    # Looking for a subject with the correct modality to initialize
    sub_paths = glob(path.join(caps_dir, '*sub-*'))
    flag_initial_subject = True
    i = 0
    while flag_initial_subject:
        sub_path = sub_paths[i]
        i += 1
        ses_paths = glob(path.join(sub_path, '*ses-*'))
        for ses_path in ses_paths:
            possible_mods = os.listdir(ses_path)
            if 'pet' in possible_mods:
                flag_initial_subject = False
                ses_init_path = ses_path
            elif i == len(sub_paths):
                raise InitException(pipeline_name)

    # Initialization must take into account the different groups !
    init_path = ses_init_path + os.sep + pet_path
    if group_selection is None:
        group_paths = glob(path.join(init_path, '*group*'))
    else:
        group_paths = []
        for group_selected in group_selection:
            new_group_path = glob(path.join(init_path, '*' + group_selected + '*'))
            group_paths = group_paths + new_group_path
            if len(new_group_path) == 0:
                warnings.warn('The group wanted does not exist for ' + pipeline_name + ': ' + group_selected, UserWarning)
    for group_path in group_paths:
        group_id = group_path.split(os.sep)[-1]
        group_list.append(group_id)
        stats_path = group_path + os.sep + 'atlas_statistics'
        # Selection of the atlas(es)
        if atlas_selection is None:
            atlas_paths = glob(path.join(stats_path, '*statistics*'))
        else:
            atlas_paths = []
            for atlas_selected in atlas_selection:
                new_atlas_path = glob(path.join(stats_path, '*' + atlas_selected + '*'))
                atlas_paths = atlas_paths + new_atlas_path
                if len(new_atlas_path) == 0:
                    warnings.warn('The atlas wanted does not exist for ' + pipeline_name + ': ' + atlas_selected + ' in group ' + group_id, UserWarning)

        if len(atlas_paths) == 0:
            raise InitException(pipeline_name)

        atlas_paths_preserved = list(atlas_paths)
        # Selection depending on the label 'pvc-rbv'
        for i, atlas_path in enumerate(atlas_paths):
            atlas_split = atlas_path.split(os.sep)[-1].split('_')
            if pvc_restriction == 1 and 'pvc-rbv' != atlas_split[6]:
                atlas_paths_preserved.remove(atlas_path)
            elif pvc_restriction == 0 and 'pvc-rbv' == atlas_split[6]:
                atlas_paths_preserved.remove(atlas_path)

        atlas_paths = atlas_paths_preserved
        atlas_paths.sort()
        for atlas_path in atlas_paths:
            row_summary_df = pd.DataFrame(index=np.arange(1), columns=columns)
            atlas_file = atlas_path.split(os.sep)[-1]
            atlas_id = '_'.join(atlas_file.split('_')[2:-1])  # We get rid of the subject and the session in the label

            init_df = pd.read_csv(atlas_path, sep='\t')
            n_regions = len(init_df['label_name'].tolist())
            label_list = [group_id + '_' + atlas_id + '_ROI-' + str(x) for x in range(n_regions)]
            for label in label_list:
                col_list.append(label)

            row_summary = [pipeline_name, group_id, atlas_id, n_regions, label_list[0], label_list[-1]]
            row_summary_df.iloc[0] = row_summary
            summary_df = pd.concat([summary_df, row_summary_df])

    # Initial empty DataFrame for the pipeline
    number_sessions = len(df)
    pipeline_df = pd.DataFrame(index=np.arange(number_sessions), columns=col_list)

    for i in range(number_sessions):
        # Reading the reference (subject + session)
        participant_id = df['participant_id'][i]
        session_id = df['session_id'][i]

        ses_path = caps_dir + os.sep + participant_id + os.sep + session_id
        mod_path = ses_path + os.sep + pet_path
        ses_df = pd.DataFrame(index=np.arange(1), columns=col_list)

        if os.path.exists(mod_path):
            # Looking for groups
            for group in group_list:
                group_path = mod_path + os.sep + group
                if os.path.exists(group_path):
                    # Looking for atlases
                    atlas_list = summary_df[summary_df.group_id == group]['atlas_id'].to_numpy()
                    for atlas in atlas_list:
                        atlas_path = group_path + os.sep + 'atlas_statistics' + os.sep
                        atlas_path = atlas_path + participant_id + '_' + session_id + '_' + atlas + '_statistics.tsv'
                        if os.path.exists(atlas_path):
                            n_regions = summary_df[(summary_df.group_id == group) &
                                                   (summary_df.atlas_id == atlas)]['regions_number'].values[0]
                            atlas_df = pd.read_csv(atlas_path, sep='\t')
                            label_list = [group + '_' + atlas + '_ROI-' + str(x) for x in
                                          range(n_regions)]
                            ses_df.loc[:, label_list] = atlas_df['mean_scalar'].to_numpy()

        pipeline_df.iloc[i] = ses_df.iloc[0]

    final_df = pd.concat([df, pipeline_df], axis=1)

    return final_df, summary_df


def t1_volume_pipeline(caps_dir, df, **kwargs):
    """
    This method merge the data of the t1-volume pipeline to the merged file containing the
    BIDS information.

    Args:
        caps_dir: the path to the CAPS directory
        df: the DataFrame containing the BIDS information
        atlas_selection: allows to choose the atlas to merge. If None all atlases are selected.

    Returns:
        final_df: a DataFrame containing the information of the bids and the pipeline
    """
    from ...pipelines.t1_volume.t1_volume_cli import T1VolumeCLI
    instance = T1VolumeCLI()
    pipeline_name = instance.name
    del instance

    # Recovering optional arguments
    if 'atlas_selection' in kwargs.keys():
        atlas_selection = kwargs['atlas_selection']
    else:
        atlas_selection = None

    if 'group_selection' in kwargs.keys():
        group_selection = kwargs['group_selection']
    else:
        group_selection = None

    t1_spm_path = 't1' + os.sep + 'spm' + os.sep + 'dartel'
    caps_dir = caps_dir + os.sep + 'subjects'

    # Lists to be sure of the column in case of missing data
    columns = ['pipeline_name', 'group_id', 'atlas_id', 'regions_number', 'first_column_name', 'last_column_name']
    summary_df = pd.DataFrame(columns=columns)
    col_list = []
    group_list = []

    # Looking for a subject with the correct modality to initialize
    sub_paths = glob(path.join(caps_dir, '*sub-*'))
    flag_initial_subject = True
    i = 0
    while flag_initial_subject:
        sub_path = sub_paths[i]
        i += 1
        ses_paths = glob(path.join(sub_path, '*ses-*'))
        for ses_path in ses_paths:
            possible_mods = os.listdir(ses_path)
            if 't1' in possible_mods:
                mod_path = path.join(ses_path, 't1')
                possible_mods = os.listdir(mod_path)
                if 'spm' in possible_mods:
                    flag_initial_subject = False
                    ses_init_path = ses_path
            elif i == len(sub_paths):
                raise InitException(pipeline_name)

    # Initialization must take into account the different groups !
    init_path = ses_init_path + os.sep + t1_spm_path
    if group_selection is None:
        group_paths = glob(path.join(init_path, '*group*'))
    else:
        group_paths = []
        for group_selected in group_selection:
            new_group_path = glob(path.join(init_path, '*' + group_selected + '*'))
            group_paths = group_paths + new_group_path
            if len(new_group_path) == 0:
                warnings.warn('The group wanted does not exist for ' + pipeline_name + ': ' + group_selected, UserWarning)
    for group_path in group_paths:
        group_id = group_path.split(os.sep)[-1]
        group_list.append(group_id)
        stats_path = group_path + os.sep + 'atlas_statistics'
        # Selection of the atlas(es)
        if atlas_selection is None:
            atlas_paths = glob(path.join(stats_path, '*statistics*'))
        else:
            atlas_paths = []
            for atlas_selected in atlas_selection:
                new_atlas_path = glob(path.join(stats_path, '*' + atlas_selected + '*'))
                atlas_paths = atlas_paths + new_atlas_path
                if len(new_atlas_path) == 0:
                    warnings.warn('The atlas wanted does not exist for ' + pipeline_name + ': ' + atlas_selected + ' in group ' + group_id, UserWarning)

        if len(atlas_paths) == 0:
            raise InitException(pipeline_name)

        atlas_paths.sort()
        for atlas_path in atlas_paths:
            row_summary_df = pd.DataFrame(index=np.arange(1), columns=columns)
            atlas_file = atlas_path.split(os.sep)[-1]
            atlas_id = '_'.join(atlas_file.split('_')[2:-1])  # We get rid of the subject and the session in the label

            init_df = pd.read_csv(atlas_path, sep='\t')
            n_regions = len(init_df['label_name'].tolist())
            label_list = [group_id + '_' + atlas_id + '_ROI-' + str(x) for x in range(n_regions)]
            for label in label_list:
                col_list.append(label)

            row_summary = [pipeline_name, group_id, atlas_id, n_regions, label_list[0], label_list[-1]]
            row_summary_df.iloc[0] = row_summary
            summary_df = pd.concat([summary_df, row_summary_df])

    # Initial empty DataFrame for the pipeline
    number_sessions = len(df)
    pipeline_df = pd.DataFrame(index=np.arange(number_sessions), columns=col_list)

    for i in range(number_sessions):
        # Reading the reference (subject + session)
        participant_id = df['participant_id'][i]
        session_id = df['session_id'][i]

        ses_path = caps_dir + os.sep + participant_id + os.sep + session_id
        mod_path = ses_path + os.sep + t1_spm_path
        ses_df = pd.DataFrame(index=np.arange(1), columns=col_list)

        if os.path.exists(mod_path):
            # Looking for groups
            for group in group_list:
                group_path = mod_path + os.sep + group
                if os.path.exists(group_path):
                    # Looking for atlases
                    atlas_list = summary_df[summary_df.group_id == group]['atlas_id'].to_numpy()
                    for atlas in atlas_list:
                        atlas_path = group_path + os.sep + 'atlas_statistics' + os.sep
                        atlas_path = atlas_path + participant_id + '_' + session_id + '_' + atlas + '_statistics.tsv'
                        if os.path.exists(atlas_path):
                            n_regions = summary_df[(summary_df.group_id == group) &
                                                   (summary_df.atlas_id == atlas)]['regions_number'].values[0]
                            atlas_df = pd.read_csv(atlas_path, sep='\t')
                            label_list = [group + '_' + atlas + '_ROI-' + str(x) for x in
                                          range(n_regions)]
                            ses_df.loc[:, label_list] = atlas_df['mean_scalar'].to_numpy()

        pipeline_df.iloc[i] = ses_df.iloc[0]

    final_df = pd.concat([df, pipeline_df], axis=1)

    return final_df, summary_df


class InitException(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return repr('The following pipeline cannot be initialized: ' + self.name)


class DatasetError(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return repr('Bad format for the sessions: ' + self.name)



# -*- coding: utf-8 -*-


import clinica.pipelines.engine as cpe


# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
from nipype import config
cfg = dict(execution={'parameterize_dirs': False})
config.update_config(cfg)


class DeepLearningPrepareData(cpe.Pipeline):
    """Deeplearning prepare data - MRI in nifti format are transformed into
    PyTorch tensors. The transformation is applied to: the whole volume, a
    selection of 3D patches, or slices extracted from the 3D volume. By default
    it uses the cropped version of the MRI (see option "--use_uncropper_image")


    Returns:
        A clinica pipeline object containing the Deeplearning prepare data pipeline.

    Raises:

    """

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """

        return ['input_nifti']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """

        return ['image_id']  # Fill here the list

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.utils.exceptions import ClinicaBIDSError, ClinicaException
        from clinica.utils.stream import cprint
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import T1W_LINEAR
        from clinica.utils.input_files import T1W_LINEAR_CROPPED
        from clinica.utils.ux import print_images_to_process

        if self.parameters.get('use_uncropped_image'):
            FILE_TYPE = T1W_LINEAR
        else:
            FILE_TYPE = T1W_LINEAR_CROPPED

        # T1w_Linear file:
        try:
            t1w_files = clinica_file_reader(self.subjects,
                                            self.sessions,
                                            self.caps_directory,
                                            FILE_TYPE)
        except ClinicaException as e:
            err = 'Clinica faced error(s) while trying to read files in your CAPS directory.\n' + str(e)
            raise ClinicaBIDSError(err)

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint('The pipeline will last approximately 30 seconds per image.')  # Replace by adequate computational time.

        if self.parameters.get('extract_method') == 'slice':
            self.slice_direction = self.parameters.get('slice_direction')
            self.slice_mode = self.parameters.get('slice_mode')
        else:
            self.slice_direction = 'axial'
            self.slice_mode = 'rgb'

        if self.parameters.get('extract_method') == 'patch':
            self.patch_size = self.parameters.get('patch_size')
            self.stride_size = self.parameters.get('stride_size')
        else:
            self.patch_size = 50
            self.stride_size = 50

        # The reading node
        # -------------------------
        read_node = npe.Node(name="ReadingFiles",
                             iterables=[
                                 ('input_nifti', t1w_files),
                             ],
                             synchronize=True,
                             interface=nutil.IdentityInterface(
                                 fields=self.get_input_fields())
                             )

        self.connect([
            (read_node, self.input_node, [('input_nifti', 'input_nifti')]),
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.utility as nutil
        from nipype.interfaces.io import DataSink
        import nipype.pipeline.engine as npe
        from clinica.utils.nipype import (fix_join, container_from_filename)
        from clinica.utils.filemanip import get_subject_id

        # Write node
        # ----------------------
        write_node = npe.Node(
                name="WriteCaps",
                interface=DataSink()
                )
        write_node.inputs.base_directory = self.caps_directory
        write_node.inputs.parameterization = False

        # Get subject ID node
        # ----------------------
        image_id_node = npe.Node(
                interface=nutil.Function(
                    input_names=['bids_or_caps_file'],
                    output_names=['image_id'],
                    function=get_subject_id),
                name='ImageID'
                )

        # Find container path from t1w filename
        # ----------------------
        container_path = npe.Node(
                nutil.Function(
                    input_names=['bids_or_caps_filename'],
                    output_names=['container'],
                    function=container_from_filename),
                name='ContainerPath')

        self.connect([
            (self.input_node, image_id_node, [('input_nifti', 'bids_or_caps_file')]),
            (self.input_node, container_path, [('input_nifti', 'bids_or_caps_filename')]),
            # (image_id_node, write_node, [('image_id', '@image_id')]),
            (image_id_node, write_node, [('image_id', '@image_id')]),
            ])

        subfolder = 'image_based'
        if self.parameters.get('extract_method') == 'slice':
            subfolder = 'slice_based'
            self.connect([
                (self.output_node, write_node, [('slices_rgb_T1', '@slices_rgb_T1')]),
                (self.output_node, write_node, [('slices_original_T1', '@slices_original_T1')])
                ])

        elif self.parameters.get('extract_method') == 'patch':
            subfolder = 'patch_based'
            self.connect([
                (self.output_node, write_node, [('patches_T1', '@patches_T1')])
                ])
        else:
            self.connect([
                (self.output_node, write_node, [('output_pt_file', '@output_pt_file')])
                ])

        self.connect([
            (container_path, write_node, [(
                (
                    'container', fix_join,
                    'deeplearning_prepare_data', subfolder, 't1_linear'
                    ),
                'container')]),
            ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from .deeplearning_prepare_data_utils import (extract_slices,
                                                      extract_patches,
                                                      save_as_pt)
        # The processing nodes

        # Node to save MRI in nii.gz format into pytorch .pt format
        # ----------------------
        save_as_pt = npe.MapNode(
               name='save_as_pt',
               iterfield=['input_img'],
               interface=nutil.Function(
                   function=save_as_pt,
                   input_names=['input_img'],
                   output_names=['output_file']
                   )
               )

        # Extract slices node (options: 3 directions, mode)
        # ----------------------
        extract_slices = npe.MapNode(
                name='extract_slices',
                iterfield=['input_tensor'],
                interface=nutil.Function(
                    function=extract_slices,
                    input_names=[
                        'input_tensor', 'slice_direction',
                        'slice_mode'
                        ],
                    output_names=['output_file_rgb', 'output_file_original']
                    )
                )

        extract_slices.inputs.slice_direction = self.slice_direction
        extract_slices.inputs.slice_mode = self.slice_mode

        # Extract patches node (options, patch size and stride size)
        # ----------------------
        extract_patches = npe.MapNode(
                name='extract_patches',
                iterfield=['input_tensor'],
                interface=nutil.Function(
                    function=extract_patches,
                    input_names=['input_tensor', 'patch_size', 'stride_size'],
                    output_names=['output_patch']
                    )
                )

        extract_patches.inputs.patch_size = self.patch_size
        extract_patches.inputs.stride_size = self.stride_size

        # Connections
        # ----------------------
        self.connect([
            (self.input_node, save_as_pt, [('input_nifti', 'input_img')]),
            ])

        if self.parameters.get('extract_method') == 'slice':
            self.connect([
                (save_as_pt, extract_slices, [('output_file', 'input_tensor')]),
                (extract_slices, self.output_node, [('output_file_rgb', 'slices_rgb_T1')]),
                (extract_slices, self.output_node, [('output_file_original', 'slices_original_T1')])
                ])
        elif self.parameters.get('extract_method') == 'patch':
            self.connect([
                (save_as_pt, extract_patches, [('output_file', 'input_tensor')]),
                (extract_patches, self.output_node, [('output_patch', 'patches_T1')])
                ])
        else:
            self.connect([
                (save_as_pt, self.output_node, [('output_file', 'output_pt_file')]),
                ])



# coding: utf8


import numpy as np
from sklearn.metrics import balanced_accuracy_score, accuracy_score, classification_report


def evaluate_prediction(y, y_hat):

    true_positive = 0.0
    true_negative = 0.0
    false_positive = 0.0
    false_negative = 0.0

    tp = []
    tn = []
    fp = []
    fn = []

    for i in range(len(y)):
        if y[i] == 1:
            if y_hat[i] == 1:
                true_positive += 1
                tp.append(i)
            else:
                false_negative += 1
                fn.append(i)
        else:  # -1
            if y_hat[i] == 0:
                true_negative += 1
                tn.append(i)
            else:
                false_positive += 1
                fp.append(i)

    accuracy = (true_positive + true_negative) / (true_positive + true_negative + false_positive + false_negative)

    if (true_positive + false_negative) != 0:
        sensitivity = true_positive / (true_positive + false_negative)
    else:
        sensitivity = 0.0

    if (false_positive + true_negative) != 0:
        specificity = true_negative / (false_positive + true_negative)
    else:
        specificity = 0.0

    if (true_positive + false_positive) != 0:
        ppv = true_positive / (true_positive + false_positive)
    else:
        ppv = 0.0

    if (true_negative + false_negative) != 0:
        npv = true_negative / (true_negative + false_negative)
    else:
        npv = 0.0

    balanced_accuracy = (sensitivity + specificity) / 2

    results = {'accuracy': accuracy,
               'balanced_accuracy': balanced_accuracy,
               'sensitivity': sensitivity,
               'specificity': specificity,
               'ppv': ppv,
               'npv': npv,
               'confusion_matrix': {'tp': len(tp), 'tn': len(tn), 'fp': len(fp), 'fn': len(fn)}
               }

    return results


def gram_matrix_linear(data):
    return np.dot(data, data.transpose())


def evaluate_prediction_multiclass(y, y_hat):

    balanced_accuracy = balanced_accuracy_score(y, y_hat)
    accuracy = accuracy_score(y, y_hat)

    results = {'accuracy': accuracy,
               'balanced_accuracy': balanced_accuracy}

    return results


def metric_distribution(metric, labels, output_path, num_classes=2, metric_label='balanced accuracy'):
    """

    Distribution plots of various metrics such as balanced accuracy!

    metric is expected to be ndarray of size [num_repetitions, num_datasets]

    """
    # from __future__ import print_function, division

    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib import cm
    from matplotlib.backends.backend_pdf import PdfPages

    num_repetitions = metric.shape[0]
    num_datasets = metric.shape[1]
    assert len(labels) == num_datasets, "Differing number of features and labels!"
    method_ticks = 1.0 + np.arange(num_datasets)

    fig, ax = plt.subplots(figsize=[9, 9])
    line_coll = ax.violinplot(metric, widths=0.8, bw_method=0.2,
                              showmedians=True, showextrema=False,
                              positions=method_ticks)

    cmap = cm.get_cmap('Paired', num_datasets)
    for cc, ln in enumerate(line_coll['bodies']):
        ln.set_facecolor(cmap(cc))
        ln.set_label(labels[cc])

    plt.legend(loc=2, ncol=num_datasets)

    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.grid(axis='y', which='major')

    lower_lim = np.round(np.min([np.float64(0.9 / num_classes), metric.min()]), 3)
    upper_lim = np.round(np.max([1.01, metric.max()]), 3)
    step_tick = 0.1
    ax.set_ylim(lower_lim, upper_lim)

    ax.set_xticks(method_ticks)
    ax.set_xlim(np.min(method_ticks) - 1, np.max(method_ticks) + 1)
    ax.set_xticklabels(labels, rotation=45)  # 'vertical'

    ax.set_yticks(np.arange(lower_lim, upper_lim, step_tick))
    ax.set_yticklabels(np.arange(lower_lim, upper_lim, step_tick))
    # plt.xlabel(xlabel, fontsize=16)
    plt.ylabel(metric_label, fontsize=16)

    fig.tight_layout()

    pp1 = PdfPages(output_path + '.pdf')
    pp1.savefig()
    pp1.close()

    return
